/*
 *  Copyright 1996, 1997, 1998 Hans Reiser, see reiserfs/README for licensing and copyright details
 */


/*
 * Contains code from
 *
 *  linux/include/linux/lock.h and linux/fs/buffer.c /linux/fs/minix/fsync.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 */


/* this is to be included from fs/buffer.c. This requires adding few lines in kernel/ksyms.c */

#ifndef __KERNEL__

#include "nokernel.h"

#endif


#define CARRY_ON                0
#define SCHEDULE_OCCURRED       1


inline int  test_and_wait_on_buffer(
              struct buffer_head * p_s_bh
            ) {
  if ( buffer_locked(p_s_bh) )  {
    __wait_on_buffer(p_s_bh);
    return SCHEDULE_OCCURRED;
  }
  return CARRY_ON;
}
 

/* This is pretty much the same as the corresponding original linux
   function, it differs only in that it tracks whether schedule
   occurred. */
/* This function function calls find_buffer which looks for the buffer
   desired on the appropriate hash queue, and then waits for it to unlock. */
struct buffer_head  * reiserfs_get_hash_table(
                        kdev_t  n_dev,
                        int     n_block,
                        int     n_size,
                        int   * p_n_repeat
                      ) {
  struct buffer_head * p_s_bh;

  for ( ; ; ) {
    if ( ! (p_s_bh = find_buffer(n_dev, n_block, n_size)) )
      return NULL;
    p_s_bh->b_count++;
    *p_n_repeat |= test_and_wait_on_buffer(p_s_bh);
    if ( *p_n_repeat == CARRY_ON ||
	 (p_s_bh->b_dev == n_dev && p_s_bh->b_blocknr == n_block && p_s_bh->b_size == n_size) )
      return p_s_bh;
    p_s_bh->b_count--;
  }
}

#ifdef __KERNEL__
struct super_block * reiserfs_get_super (kdev_t dev)
{
  struct super_block * s;
  
  if (!dev)
    return NULL;

  s = sb_entry(super_blocks.next);
  while (s != sb_entry(&super_blocks)) {
    if (s->s_dev == dev)
      return s;
    else
      s = sb_entry(s->s_list.next);
  }
  return NULL;
}


static void reiserfs_end_buffer_io_sync (struct buffer_head *bh, int uptodate)
{
  struct super_block * s = reiserfs_get_super (bh->b_dev);
  if (s && s->u.reiserfs_sb.unpreserve)
    s->u.reiserfs_sb.unpreserve (s, bh);
  /*  unpreserve (reiserfs_get_super (bh->b_dev), bh);*/
  mark_buffer_uptodate(bh, uptodate);
  unlock_buffer(bh);
}



/* This is pretty much the same as the corresponding original linux
   function, it differs only in that it tracks whether schedule
   occurred. */
/* This function looks for a buffer which contains a given block.  If
   the block is in cache it returns it, otherwise it returns a new
   buffer which is not uptodate.  This is called by reiserfs_bread and
   other functions. Note that get_new_buffer ought to be called this
   and this ought to be called get_new_buffer, since this doesn't
   actually get the block off of the disk. */
struct buffer_head  * reiserfs_getblk(
                        kdev_t    n_dev,
                        int       n_block,
                        int       n_size,
                        int     * p_n_repeat
                      ) {
   struct buffer_head  * p_s_bh;
  int                   n_isize = BUFSIZE_INDEX(n_size);

repeat:

  p_s_bh = reiserfs_get_hash_table(n_dev, n_block, n_size, p_n_repeat);
  if ( p_s_bh ) {
    if ( ! buffer_dirty(p_s_bh) ) {
      if ( buffer_uptodate(p_s_bh) )
	put_last_lru(p_s_bh);
      p_s_bh->b_flushtime = 0;
    }
    return p_s_bh;
  }
  
  while ( ! free_list[n_isize] )  {
    *p_n_repeat |= SCHEDULE_OCCURRED;
    refill_freelist(n_size);
  }
	
  if ( *p_n_repeat != CARRY_ON && find_buffer(n_dev, n_block, n_size) )
    goto repeat;

  p_s_bh = free_list[n_isize];
  remove_from_free_list(p_s_bh);

  /* OK, FINALLY we know that this buffer is the only one of its kind,
   * and that it's unused (b_count=0), unlocked (buffer_locked=0), and clean.
   */
  init_buffer(p_s_bh, n_dev, n_block, reiserfs_end_buffer_io_sync, NULL);
  p_s_bh->b_state=0;
  insert_into_queues(p_s_bh);
  return p_s_bh;
}



static void print_buffer_head2 (struct buffer_head * bh)
{
  if (bh == 0) {
    printk ("print_buffer_head2: 0\n");
    return;
  }
  
  printk ("b_data=%p, b_blocknr = %lu, ", bh->b_data, bh->b_blocknr);
  printk ("b_dev=0x%x, ", bh->b_dev);
  printk ("b_state = 0x%lx, ", bh->b_state);
  printk ("b_count = %d, ", bh->b_count);
  printk ("b_size = %lu\n", bh->b_size);
  
}


void reiserfs_show_buffers (kdev_t dev)
{
	struct buffer_head * bh;
	int found = 0, locked = 0, dirty = 0, used = 0, lastused = 0;
	int protected = 0;
	int nlist;
	/*
	static char *buf_types[NR_LIST] = {"CLEAN","LOCKED","DIRTY"};

	printk("Buffer memory:   %6dkB\n",buffermem>>10);
	printk("Buffer heads:    %6d\n",nr_buffer_heads);
	printk("Buffer blocks:   %6d\n",nr_buffers);
	*/
	for(nlist = 0; nlist < NR_LIST; nlist++) {
	  found = locked = dirty = used = lastused = protected = 0;
	  bh = lru_list[nlist];
	  if(!bh) continue;

	  do {
            	if (bh->b_dev == dev  && bh->b_count != 0) {
		  	print_buffer_head2 (bh);
		}

		found++;
		if (buffer_locked(bh))
			locked++;
		if (buffer_protected(bh))
			protected++;
		if (buffer_dirty(bh))
			dirty++;
		if (bh->b_count)
			used++, lastused = found;
		bh = bh->b_next_free;
	  } while (bh != lru_list[nlist]);
	  /*
	  printk("%8s: %d buffers, %d used (last=%d), "
		 "%d locked, %d protected, %d dirty\n",
		 buf_types[nlist], found, used, lastused,
		 locked, protected, dirty);
		 */
	};

}


void mark_suspected_recipients_dirty (kdev_t dev)
{
  struct buffer_head * bh;
  int nlist;
  int dirtied = 0;

repeat:
  for(nlist = 0; nlist < NR_LIST; nlist++) {
    bh = lru_list[nlist];
    if(!bh) continue;
    
    do {
      if (bh->b_dev == dev && test_bit (12, &(bh->b_state))) {
	if (!buffer_dirty (bh)) {
	  mark_buffer_dirty (bh, 1);
	  dirtied ++;
	  goto repeat;
	}
      }
      bh = bh->b_next_free;
    } while (bh != lru_list[nlist]);
  }
  if (dirtied)
    printk ("mark_suspected_recipients_dirty: found %d suspected recipients, dirtied\n", dirtied);
}



void invalidate_reiserfs_buffers (kdev_t dev)
{
  int i;
  int nlist;
  struct buffer_head * bh;
  
  for(nlist = 0; nlist < NR_LIST; nlist++) {
    bh = lru_list[nlist];
    for (i = nr_buffers_type[nlist]*2 ; --i > 0 ; bh = bh->b_next_free) {
      if (bh->b_dev != dev)
	continue;
      wait_on_buffer(bh);
      if (bh->b_dev != dev)
	continue;
      /*if (bh->b_count)
	continue;*/
      
      bh->b_end_io = end_buffer_io_sync;
    }
  }
}

#endif /* __KERNEL__ */
