/*
 * Copyright 1996, 1997, 1998 Hans Reiser, see reiserfs/README for licensing and copyright details
 */
#ifdef __KERNEL__

#include <linux/sched.h>
#include <linux/reiserfs_fs.h>

#else

#include "nokernel.h"

#endif


/*
  symlink, link, mknod, rename are not ready
*/



static int compare_names (struct buffer_head * bh, struct item_head * ih, const char * name, int namelen)
{
  int retval = ENTRY_NOT_FOUND;
  
  if ((namelen == B_I_NAMELEN (bh, ih)) && !memcmp (B_I_NAME(bh,ih), name, namelen))
    retval = is_entry_hidden (ih) ? ENTRY_FOUND_HIDDEN : ENTRY_FOUND;

  return retval;
}


/* this extracts object's stat data key from the directory entry */
static void get_object_key (struct key * key, struct buffer_head * bh, struct item_head * ih)
{
  struct key * key_in_entry;

  key->k_offset = 0;
  key->group_locale.minor = 0;
  if (is_entry_pointing_to_dir (ih)) {
    key->group_locale.major = key->file_id = ih->ih_key.file_id;
    key->name_hash = 0;
    return;
  }
  key_in_entry = B_I_OBJECT_LOCALE_FROM_ENTRY (bh, ih);
  key->group_locale.major = key_in_entry->group_locale.major;
  key->name_hash = le32_to_cpu (key_in_entry->name_hash);
  key->file_id = le32_to_cpu (key_in_entry->file_id);
}


/* starting from item, pointed by ih, check every entry-items if they have given hash value */
static int scan_items_to_left (struct path * path, struct item_head ** ih,
			       __u32 hash, const char * name, int namelen, struct key * object_key)
{
  struct buffer_head * bh = PATH_PLAST_BUFFER (path);

  while (1) {
    if (is_stat_data (*ih) || hash != (*ih)->ih_key.name_hash)
      /* stop scanning as we have reached a stat data of the directory we looked through or hash
	 does not match to the given one */
      return ENTRY_NOT_FOUND;

    if (is_dir_entry (*ih) && compare_names (bh, *ih, name, namelen) == ENTRY_FOUND) {
      get_object_key (object_key, bh, *ih);
      return ENTRY_FOUND;
    }

    if (*ih == B_N_PITEM_HEAD (bh, 0))
      break;

    (*ih)--;
    PATH_LAST_POSITION (path)--;
  }

  return GOTO_PREVIOUS_ITEM;
}


static int name_is_dot_or_dot_dot (struct inode * dir, const char * name, int namelen, struct key * object_key)
{
  if (!namelen || (namelen == 1 && name[0] == '.')) {
    object_key->group_locale.major = object_key->file_id = dir->i_ino;
    object_key->group_locale.minor = 0;
    object_key->name_hash = object_key->k_offset = 0;
    return 1;
  }
  if (namelen == 2 && name[0] == '.' && name[1] == '.') {
    object_key->group_locale.major = object_key->file_id = INODE_DOT_DOT_GROUP_LOCALE (dir);
    object_key->group_locale.minor = 0;
    object_key->name_hash = object_key->k_offset = 0;
    return 1;
  }

  return 0;
}


static int reiserfs_find_entry (struct inode * dir, const char * name, int namelen, 
				struct path * path, struct key * object_key)
{
  struct key key_to_search;
  struct item_head * ih;
  int repeat;
  int retval;

  if (!dir || !dir->i_sb)
    return ENTRY_NOT_FOUND;

  if (namelen > REISERFS_MAX_NAME_LEN (dir->i_sb->s_blocksize))
    return ENTRY_NOT_FOUND;

  /* there are no items for "." and ".." */
  if (name_is_dot_or_dot_dot (dir, name, namelen, object_key))
    return ENTRY_FOUND;

  /* compose key to search so, that we will find the last item among all ones having the same value
     of name_hash key field */
  key_to_search.group_locale.major = dir->i_ino; /* i_ino is unsigned long, pack_locale is __u32 */
  key_to_search.group_locale.minor = 0;
  key_to_search.name_hash = keyed_hash (name, namelen) & 0x7fffffff;
  key_to_search.file_id = (__u32)-1;
  key_to_search.k_offset = 0;

  while (1) {
    /* search for a directory item using the formed key */
    if (search_by_key (dir->i_sb, &key_to_search, path, &repeat, DISK_LEAF_NODE_LEVEL, READ_BLOCKS) == ITEM_NOT_FOUND) {
      /* take previous item */
      if (PATH_LAST_POSITION (path) == 0)
	reiserfs_panic (dir->i_sb, "vs-7015: reiserfs_find_entry: search_by_key returned position 0 for key %k",
			&key_to_search);
      PATH_LAST_POSITION (path) --;
    }

    ih = PATH_PITEM_HEAD (path);

#ifdef REISERFS_CHECK

    /* make sure, that we get into desired directory */
    if (ih->ih_key.group_locale.major != dir->i_ino)
      reiserfs_panic (dir->i_sb, "vs-7020: reiserfs_find_entry: looking for %k we get near item %h",
		      &key_to_search, ih);

#endif /* REISERFS_CHECK */

    /* compare names for all entries having given hash value */
    retval = scan_items_to_left (path, &ih, key_to_search.name_hash, name, namelen, object_key);
    if (retval != GOTO_PREVIOUS_ITEM) {
      /* there is no need to scan directory anymore. Given entry found or does not exist */
      return retval;
    }
    
    /* directory could continue to the left. Go there */
    copy_key (&key_to_search, &(ih->ih_key));
    decrement_key (&key_to_search);
    pathrelse (path);

  } /* while (1) */
}


#define FROM_CREATE 0
#define FROM_MKDIR 1
#define FROM_RENAME 2
#define FROM_LINK 3

static int get_entry_item_len (int namelen, int where)
{
  if (where == FROM_CREATE)
    return namelen + SD_SIZE;
  if (where == FROM_MKDIR)
    return namelen;
  return 0;
}


/* allocate space for item body of the new entry, set locality, create stat data if needed, copy name */
static char * compose_entry_item_body (const char * name, int namelen, struct inode * inode, struct item_head * ih, int where)
{
  char * buf, * p;

  buf = (char *)kmalloc (ih->ih_item_len, GFP_KERNEL);
  if (buf == 0)
    return 0;

  p = buf;

  if (where == FROM_CREATE) {
    /* regular file or symbolic link is being created. Copy stat data to the item */
    inode2sd ((struct stat_data *)p, inode);
    ((struct stat_data *)p)->u.first_direct_byte = cpu_to_le32 (S_ISLNK (inode->i_mode) ? 1 : NO_BYTES_IN_DIRECT_ITEM);
    mark_entry_with_stat_data (ih);
    INODE_PKEY (inode)->name_hash = ih->ih_key.name_hash;
    p += SD_SIZE;
  } else if (where == FROM_MKDIR) {
    mark_entry_points_to_dir (ih);
  } else if (where == FROM_LINK) {
    /* set key of the object, new entry will point to */
    copy_object_locale ((struct key *)p, INODE_PKEY (inode));
    p += OBJECT_LOCALE_SIZE;
  }

  /* copy name to the entry item */
  memcpy (p, name, namelen);

  return buf;
}


/* directory entry is a separate item. If inode is of a regular file and its nlink == 1, stat data
   will get into entry. */
static int reiserfs_add_entry (struct inode * dir, const char * name, int namelen, struct inode * inode, int where)
{
  struct item_head ih;
  char * buf;
  struct path path;
  int repeat;

  init_path (&path);

  if (!dir || !dir->i_sb)
    return -ENOENT;

  if (namelen > REISERFS_MAX_NAME_LEN (dir->i_sb->s_blocksize))
    return -ENAMETOOLONG;

  /* prepare item header */
  ih.ih_key.group_locale.major = dir->i_ino;
  ih.ih_key.group_locale.minor = 0;
  ih.ih_key.name_hash = keyed_hash (name, namelen) & 0x7fffffff;
  ih.ih_key.file_id = inode->i_ino;
  ih.ih_key.k_offset = 0;
  mark_dir_entry (&ih);
  if (where == FROM_RENAME)
    mark_entry_hidden (&ih);
  ih.ih_item_len = get_entry_item_len (namelen, where);
  ih.ih_free_space = 0;

  /* prepare item body */
  buf = compose_entry_item_body (name, namelen, inode, &ih, where);
  if (buf == 0)
    return -ENOMEM;

  /* find the proper place for the new entry */
  if (search_by_key (dir->i_sb, &(ih.ih_key), &path, &repeat, DISK_LEAF_NODE_LEVEL, READ_BLOCKS) == ITEM_FOUND)
    reiserfs_panic (dir->i_sb, "reiserfs_add_entry: 7032: item with this key %k already exists %h", &(ih.ih_key));      
  
  /* perform the insertion of the entry that we have prepared */
  if (reiserfs_insert_item (dir->i_sb, &path, &ih, buf, 0/*zeros_num*/, NOTHING_SPECIAL) != OK) {
    kfree (buf);
    return -ENOSPC;
  }

  kfree (buf);
  dir->i_size += (ih.ih_item_len + IH_SIZE);
  dir->i_mtime = dir->i_ctime = CURRENT_TIME;
  /* note, that dir is not marked as dirty */
  return 0;
}


int reiserfs_lookup (struct inode * dir, struct dentry * dentry)
{
  struct inode * inode = 0;
  struct path path_to_entry;
  struct key object_key;

  init_path (&path_to_entry);

  if (reiserfs_find_entry (dir, dentry->d_name.name, dentry->d_name.len, &path_to_entry, &object_key) == ENTRY_FOUND) {
    inode = reiserfs_iget (dir->i_sb, &path_to_entry, &object_key);
    if (!inode)
      return -EACCES;
  } else
    pathrelse (&path_to_entry);

  d_add(dentry, inode);
  return 0;
}


int reiserfs_create (struct inode * dir, struct dentry * dentry, int mode)
{
  int error;
  struct inode * inode;
	
  if (!dir)
    return -ENOENT;

  inode = reiserfs_new_inode (dir);
  if (!inode) {
    return -ENOSPC;
  }

  inode->i_op = &reiserfs_file_inode_operations;
  inode->i_mode = mode;

  error = reiserfs_add_entry (dir, dentry->d_name.name, dentry->d_name.len, inode, FROM_CREATE);
  if (error) {
    inode->i_nlink--;
    if_in_ram_update_sd (inode);
    iput (inode);
    return error;
  }
  if_in_ram_update_sd (dir);
  d_instantiate(dentry, inode);
  return 0;
}


int reiserfs_mkdir (struct inode * dir, struct dentry * dentry, int mode)
{
  int error;
  struct inode * inode;
  struct key object_key;
  struct path path_to_entry;

  init_path (&path_to_entry);

  if (!dir || !dir->i_sb) {
    return -EINVAL;
  }

  if (reiserfs_find_entry (dir, dentry->d_name.name, dentry->d_name.len, &path_to_entry, &object_key) == ENTRY_FOUND) {
    pathrelse (&path_to_entry);
    return -EEXIST;
  }
  pathrelse (&path_to_entry);
  
  if (dir->i_nlink >= REISERFS_LINK_MAX) {
    return -EMLINK;
  }
  /*
  mode = S_IFDIR | (mode & 0777 & ~current->fs->umask);
  if (dir->i_mode & S_ISGID)
    mode |= S_ISGID;*/
  inode = reiserfs_new_directory_inode (dir, mode);
  if (!inode) {
    return -ENOSPC;
  }

  inode->i_op = &reiserfs_dir_inode_operations;

  /* new inode and stat data are uptodate. Inode is clean. */
  error = reiserfs_add_entry (dir, dentry->d_name.name, dentry->d_name.len, inode, FROM_MKDIR);
  if (error) {
    inode->i_nlink = 0;
    if_in_ram_update_sd (inode);
    iput (inode);
    return error;
  }

  /* update dir inode, reiserfs_add_entry does not do that */
  dir->i_nlink ++;
  if_in_ram_update_sd (dir);
  d_instantiate(dentry, inode);
  return 0;
}


static int reiserfs_empty_dir(struct inode * inode)
{
  return inode->i_size == EMPTY_DIR_SIZE;
}


int rmdir_not_allowed (struct inode * dir, struct key * key_in_entry, struct inode * inode)
{
  if ((dir->i_mode & S_ISVTX)/* && !fsuser() &&
      current->fsuid != inode->i_uid &&
      current->fsuid != dir->i_uid*/)
    return -EPERM;

  if (inode->i_dev != dir->i_dev)
    return -EPERM;

  if (inode == dir)	/* we may not delete ".", but "../dir" is ok */
    return -EPERM;

  if (!S_ISDIR(inode->i_mode))
    return -ENOTDIR;

  if (!reiserfs_empty_dir(inode))
    return -ENOTEMPTY;

  if (key_in_entry->file_id != inode->i_ino)
    return -ENOENT;

  return 0;
}


int reiserfs_rmdir (struct inode * dir, struct dentry *dentry)
{
  struct inode * inode;
  int retval;
  struct key key_in_entry;
  struct item_head entry_item;
  struct path path;

  init_path (&path);

  retval = -ENOENT;
  if (reiserfs_find_entry (dir, dentry->d_name.name, dentry->d_name.len, &path, &key_in_entry) == ENTRY_NOT_FOUND)
    goto end_rmdir;

  inode = dentry->d_inode;

  retval = rmdir_not_allowed (dir, &key_in_entry, inode);
  if (retval)
    goto end_rmdir;

  copy_item_head (&entry_item, PATH_PITEM_HEAD (&path));
     
  down (&inode->i_sem);

  /* free preserve list if we should */
  maybe_free_preserve_list (dir->i_sb);

  if (!reiserfs_empty_dir (inode))
    retval = -ENOTEMPTY;
  else {
    /* cut entry from dir directory (delete corresponding entry item) */
    if (reiserfs_delete_item (dir->i_sb, &path, 0/*pos_in_item*/, &(entry_item.ih_key), 0/*unfm_bh*/, NOTHING_SPECIAL) != entry_item.ih_item_len) {
      retval = -ENOENT;
    }
  }
  up(&inode->i_sem);
  if (retval)
    goto end_rmdir;

  if (inode->i_nlink != 2)
    printk ("reiserfs_rmdir: empty directory has nlink != 2 (%d)\n", inode->i_nlink);
  inode->i_nlink = 0;
  inode->i_ctime = dir->i_ctime = dir->i_mtime = CURRENT_TIME;
  if_in_ram_update_sd (inode);
  dir->i_nlink --;
  dir->i_size -= (IH_SIZE + entry_item.ih_item_len);
  if_in_ram_update_sd (dir);

  d_delete(dentry);
  return 0;
	
end_rmdir:
  /* we must release path, because we did not call reiserfs_cut_from_item, or reiserfs_cut_from_item
     does not release path if operation was not complete */
  pathrelse (&path);
  return retval;	
}


int reiserfs_unlink (struct inode * dir, struct dentry *dentry)
{
  int retval = -ENOENT;
  struct inode * inode;
  struct key key_in_entry;
  struct item_head entry_item;
  struct path path;

  init_path (&path);

  /* free preserve list if we should */
  maybe_free_preserve_list (dir->i_sb);
	
  if (reiserfs_find_entry (dir, dentry->d_name.name, dentry->d_name.len, &path, &key_in_entry) == ENTRY_NOT_FOUND) {
    goto end_unlink;
  }

  inode = dentry->d_inode;

  retval = -EPERM;
  if (S_ISDIR (inode->i_mode)) {
    goto end_unlink;
  }
  if ((dir->i_mode & S_ISVTX)/* && !fsuser() &&
      current->fsuid != inode->i_uid &&
      current->fsuid != dir->i_uid*/) {
    goto end_unlink;
  }

  retval = -EIO;
  if (key_in_entry.file_id != inode->i_ino) {
    goto end_unlink;
  }
  
  if (!inode->i_nlink) {
    printk("reiserfs_unlink: deleting nonexistent file (%s:%lu), %d\n",
	   kdevname(inode->i_dev), inode->i_ino, inode->i_nlink);
    inode->i_nlink = 1;
  }

  /* delete dir entry item. Its length can not be changed */
  copy_item_head (&entry_item, PATH_PITEM_HEAD (&path));
  if (reiserfs_delete_item (dir->i_sb, &path, 0/*pos_in_item*/, &(entry_item.ih_key), 0/*unfm_bh*/, NOTHING_SPECIAL) != entry_item.ih_item_len) {
printk ("reiserfs_delete_item returned != %d\n", entry_item.ih_item_len);
    retval = -ENOENT;
    goto end_unlink;
  }

  inode->i_nlink--;
  inode->i_ctime = CURRENT_TIME;
  if_in_ram_update_sd (inode);

  dir->i_size -= (entry_item.ih_item_len + IH_SIZE);
  dir->i_ctime = dir->i_mtime = CURRENT_TIME;
  if_in_ram_update_sd (dir);
  d_delete(dentry);
  retval = 0;

end_unlink:
  pathrelse (&path);
  return retval;
}

#if 0
int reiserfs_symlink (struct inode * dir, struct dentry * dentry, const char * symname)
{
  struct inode * inode;
  struct key object_key;
  struct path path_to_entry;
  int error;

  init_path (&path_to_entry);

  if (strlen (symname) > MAX_ITEM_LEN (dir->i_sb->s_blocksize)) {
    return -ENAMETOOLONG;
  }

  if (reiserfs_find_entry (dir, dentry->d_name.name, dentry->d_name.len, &path_to_entry, &object_key) == ENTRY_FOUND) {
    pathrelse (&path_to_entry);
    return -EEXIST;
  }

  /* insert body of link as direct item */
  inode = reiserfs_new_symlink_inode (dir, symname);
  if (!inode) {
    return -ENOSPC;
  }
  inode->i_op = &reiserfs_symlink_inode_operations;
  inode->i_size = strlen (symname);
  inode->i_mode = S_IFLNK | 0777;

  /* add entry together with stat data */
  error = reiserfs_add_entry (dir, dentry->d_name.name, dentry->d_name.len, inode, FROM_CREATE);
  if (error) {
    inode->i_nlink--;
    if_in_ram_update_sd (inode);
    iput (inode);
    return error;
  }
  if_in_ram_update_sd (dir);
  d_instantiate(dentry, inode);
  return 0;
}


int reiserfs_mknod (struct inode * dir, struct dentry *dentry, int mode, int rdev)
{
  int error;
  struct inode * inode;
  struct reiserfs_dir_entry de;
  struct path path_to_entry;

  init_path (&path);

  if (!dir)
    return -ENOENT;

  de.de_gen_number_bit_string = 0;
  if (reiserfs_find_entry (dir, dentry->d_name.name, dentry->d_name.len, &path_to_entry, &de) == POSITION_FOUND) {
    pathrelse ( &path_to_entry);
    return -EEXIST;
  }

  pathrelse ( &path_to_entry);

  inode = reiserfs_new_inode (dir, mode, 0);
  if (!inode) {
    return -ENOSPC;
  }

  inode->i_uid = current->fsuid;
  inode->i_mode = mode;
  inode->i_op = NULL;

  if (S_ISREG(inode->i_mode))
    inode->i_op = &reiserfs_file_inode_operations;
  else if (S_ISDIR(inode->i_mode)) {
    inode->i_op = &reiserfs_dir_inode_operations;
    if (dir->i_mode & S_ISGID)
      inode->i_mode |= S_ISGID;
  }
  else if (S_ISLNK(inode->i_mode))
    inode->i_op = &reiserfs_symlink_inode_operations;
  else if (S_ISCHR(inode->i_mode))
    inode->i_op = &chrdev_inode_operations;
  else if (S_ISBLK(inode->i_mode))
    inode->i_op = &blkdev_inode_operations;
  else if (S_ISFIFO(inode->i_mode))
    init_fifo(inode);
  if (S_ISBLK(mode) || S_ISCHR(mode))
    inode->i_rdev = to_kdev_t(rdev);

  if_in_ram_update_sd (inode);

  error = reiserfs_add_entry (dir, dentry->d_name.name, dentry->d_name.len, INODE_PKEY (inode), &de, 1);
  if (error) {
    inode->i_nlink--;
    if_in_ram_update_sd (inode);
    iput (inode);
    return error;
  }
  if_in_ram_update_sd (dir);
  d_instantiate(dentry, inode);
  return 0;
}


int reiserfs_link (struct dentry * old_dentry, struct inode * dir, struct dentry * dentry)
{
  struct inode *inode = old_dentry->d_inode;
  struct reiserfs_dir_entry de;
  int error;
  struct path path_to_entry;
  
  init_path (&path_to_entry);

  /* object must not be directory */
  if (S_ISDIR(inode->i_mode)) {
    return -EPERM;
  }
  
  /* file has too many links */
  if (inode->i_nlink >= REISERFS_LINK_MAX) {
    return -EMLINK;
  }

  de.de_gen_number_bit_string = 0;
  if (reiserfs_find_entry (dir, dentry->d_name.name, dentry->d_name.len, &path_to_entry, &de) == POSITION_FOUND) {
    pathrelse (&path_to_entry);
    return -EEXIST;
  }
  
  pathrelse (&path_to_entry);

  /* free preserve list if we should */
  maybe_free_preserve_list (dir->i_sb);

  /* create new entry */
  error = reiserfs_add_entry (dir, dentry->d_name.name, dentry->d_name.len, INODE_PKEY (inode), &de, FROM_LINK);
  if (error) {
    return error;
  }
  inode->i_nlink++;
  inode->i_ctime = CURRENT_TIME;
  if_in_ram_update_sd (inode);

  if_in_ram_update_sd (dir);
  inode->i_count++;
  d_instantiate(dentry, inode);
  return 0;
}


static int de_still_valid (const char * name, int len, struct reiserfs_dir_entry * de)
{
  struct item_head * ih;
  struct reiserfs_de_head * deh;

  if (!de || !de->de_bh)
    return 0;

  deh = B_I_DEH (de->de_bh, ih = B_N_PITEM_HEAD (de->de_bh, de->de_item_num));
  /* compare dir entry headers, record loacation and names */
  if (memcmp (&(deh[de->de_entry_num]), de->de_deh, DEH_SIZE) ||
      B_I_E_NAME (de->de_entry_num, de->de_bh, ih) != de->de_name ||
      memcmp (name, de->de_name, len))
    return 0;
  return 1;
}


static int entry_points_to_object (const char * name, int len, struct reiserfs_dir_entry * de, struct inode * inode)
{
  if (!de_still_valid (name, len, de))
    return 0;

  if (inode) {
    if (!de_visible (de->de_deh))
      reiserfs_panic (0, "vs-7042: entry_points_to_object: entry must be visible");
    return (de->de_objectid == inode->i_ino) ? 1 : 0;
  }

  /* this must be added hidden entry */
  if (de_visible (de->de_deh))
    reiserfs_panic (0, "vs-7043: entry_points_to_object: entry must be visible");

  return 1;
}


/* sets key of parent directory in ".." entry */
static void set_ino_in_dir_entry (struct reiserfs_dir_entry * de, struct key * key)
{
  de->de_deh->deh_dir_id = key->k_dir_id;;
  de->de_deh->deh_objectid = key->k_objectid;
  /*
  *((unsigned long *)(de->de_name + 2)) = key->k_dir_id;
  mark_de_with_directory_id (de->de_deh);*/
}


static int do_reiserfs_rename (struct inode * old_dir, struct dentry *old_dentry,
			       struct inode * new_dir, struct dentry *new_dentry)
{
  int retval;
  struct reiserfs_dir_entry old_de, new_de, dot_dot_de;
  struct inode * old_inode, * new_inode;
  int new_entry_added = 0;
  struct path old_entry_path, new_entry_path, dot_dot_entry_path;

  init_path (&old_entry_path);
  init_path (&new_entry_path);
  init_path (&dot_dot_entry_path);

  goto start_up;

try_again:
  current->counter = 0;
  schedule();
	
start_up:
  old_inode = new_inode = NULL;
  dot_dot_de.de_bh = 0;
  new_de.de_bh = 0;

  /* look for the old name in old directory */
  retval = -ENOENT;
  old_de.de_gen_number_bit_string = 0;
  if (reiserfs_find_entry (old_dir, old_dentry->d_name.name, old_dentry->d_name.len, &old_entry_path, &old_de) == POSITION_NOT_FOUND)
    goto end_rename;

  old_inode = old_dentry->d_inode;
  retval = -EPERM;

  if ((old_dir->i_mode & S_ISVTX) && 
      current->fsuid != old_inode->i_uid &&
      current->fsuid != old_dir->i_uid && !fsuser())
    goto end_rename;

  new_inode = new_dentry->d_inode;

  /* look for the new entry in target directory */
  new_de.de_gen_number_bit_string = 0;
  if (reiserfs_find_entry (new_dir, new_dentry->d_name.name, new_dentry->d_name.len, &new_entry_path, &new_de) == POSITION_FOUND) {
    if (!new_inode) {
      printk ("do_reiserfs_rename: new entry found, inode == 0 though\n");
    }
    /* this entry already exists, we can just set key of object */
    new_entry_added = 1;
  } else {
#ifdef REISERFS_CHECK
    if (new_entry_added) {
      if (new_de.de_namelen != new_dentry->d_name.len || memcmp (new_de.de_name, new_dentry->d_name.name, new_de.de_namelen) ||
	  de_visible (new_de.de_deh))
	reiserfs_panic (old_dir->i_sb, "vs-7045: do_reiserfs_rename: suspicious entry found");
    }
#endif /* REISERFS_CHECK */
  }

  if (new_inode == old_inode) {
    retval = 0;
    goto end_rename;
  }

  if (new_inode && S_ISDIR(new_inode->i_mode)) {
    /* new name exists and points to directory */
    retval = -EISDIR;
    if (!S_ISDIR(old_inode->i_mode))
      goto end_rename;
    retval = -EINVAL;
    if (is_subdir (new_dentry, old_dentry))
      goto end_rename;
    retval = -ENOTEMPTY;
    if (!reiserfs_empty_dir (new_inode))
      goto end_rename;
    retval = -EBUSY;
    if (new_inode->i_count > 1)
      goto end_rename;
  }

  retval = -EPERM;
  if (new_inode && (new_dir->i_mode & S_ISVTX) && 
      current->fsuid != new_inode->i_uid &&
      current->fsuid != new_dir->i_uid && !fsuser())
    goto end_rename;

  if (S_ISDIR(old_inode->i_mode)) {
    /* old name points to directory */
    retval = -ENOTDIR;
    if (new_inode && !S_ISDIR(new_inode->i_mode))
      goto end_rename;

    retval = -EINVAL;
    if (is_subdir(new_dentry, old_dentry))
      goto end_rename;

    retval = -EIO;
    /* directory is renamed, its parent directory will be changed, so find ".." entry */
    dot_dot_de.de_gen_number_bit_string = 0;
    if (reiserfs_find_entry (old_inode, "..", 2, &dot_dot_entry_path, &dot_dot_de) == POSITION_NOT_FOUND)
      goto end_rename;
    if (dot_dot_de.de_objectid != old_dir->i_ino)
      goto end_rename;

    retval = -EMLINK;
    if (!new_inode && new_dir->i_nlink >= REISERFS_LINK_MAX)
      goto end_rename;
  }
  
  if (new_entry_added == 0) {
    /* add new entry if we did not do it, but do not mark it as visible */
    retval = reiserfs_add_entry (new_dir, new_dentry->d_name.name, new_dentry->d_name.len, INODE_PKEY (old_inode), &new_de, 0);
    if (retval)
      goto end_rename;
    if_in_ram_update_sd (new_dir);
    new_entry_added = 1;
    goto try_again;
  }

  /* sanity checking before doing the rename - avoid races */
  if (!entry_points_to_object (new_dentry->d_name.name, new_dentry->d_name.len, &new_de, new_inode))
    goto try_again;
  if (!entry_points_to_object (old_dentry->d_name.name, old_dentry->d_name.len, &old_de, old_inode))
    /* go to re-looking for old entry */
    goto try_again;

  if (S_ISDIR(old_inode->i_mode) && !entry_points_to_object ("..", 2, &dot_dot_de, old_dir))
    /* go to re-looking for ".." entry of renamed directory */
    goto try_again;
  
  /* ok, all the changes can be done in one fell swoop when we have
     claimed all the buffers needed.*/

  /* make old name hidden */
  mark_de_hidden (old_de.de_deh);
  mark_buffer_dirty (old_de.de_bh, 1);

  /* make new name visible and set key of old object (if entry
     existed, it is already visible, if not, key is correct already) */
  mark_de_visible (new_de.de_deh);
  new_de.de_deh->deh_dir_id = INODE_PKEY (old_inode)->k_dir_id;
  new_de.de_deh->deh_objectid = INODE_PKEY (old_inode)->k_objectid;
  mark_buffer_dirty (new_de.de_bh, 1);
  mark_suspected_recipient (old_inode->i_sb, new_de.de_bh);

  old_dir->i_ctime = old_dir->i_mtime = CURRENT_TIME;
  if_in_ram_update_sd (old_dir);

  new_dir->i_ctime = new_dir->i_mtime = CURRENT_TIME;
  if_in_ram_update_sd (new_dir);

  if (new_inode) {
    new_inode->i_nlink--;
    new_inode->i_ctime = CURRENT_TIME;
    if_in_ram_update_sd (new_inode);
  }
  if (dot_dot_de.de_bh) {
    set_ino_in_dir_entry (&dot_dot_de, INODE_PKEY (new_dir));
    mark_buffer_dirty (dot_dot_de.de_bh, 1);
    old_dir->i_nlink--;
    if_in_ram_update_sd (old_dir);
    if (new_inode) {
      new_inode->i_nlink--;
      if_in_ram_update_sd (new_inode);
    } else {
      new_dir->i_nlink++;
      if_in_ram_update_sd (new_dir);
    }
  }

  /* ok, renaming done */

  /* remove old name (it is hidden now) */
  if (reiserfs_cut_from_item (
			      old_dir, old_dir->i_sb,
			      &old_entry_path, &(old_de.de_entry_num),
			      &(old_de.de_entry_key), 0, PRESERVE_RENAMING
			      ) == 0)
    printk ("reiserfs_rename: could not remove old name\n");
  else {
    old_dir->i_size -= DEH_SIZE + old_de.de_entrylen;
    if_in_ram_update_sd (old_dir);
  }

  /* Update the dcache */
  d_move(old_dentry, new_dentry);
  retval = 0;

end_rename:
  pathrelse (&old_entry_path);
  pathrelse (&new_entry_path);
  pathrelse (&dot_dot_entry_path);
  return retval;
}


int	reiserfs_rename (
			 struct inode * old_dir, struct dentry *old_dentry,
			 struct inode * new_dir, struct dentry *new_dentry
			 )
{
  static struct wait_queue * wait = NULL;
  static int lock = 0;
  int result;
  
  while (lock)
    sleep_on(&wait);
  lock = 1;
  result = do_reiserfs_rename (old_dir, old_dentry, new_dir, new_dentry);
  lock = 0;
  wake_up(&wait);
  return result;
}



#endif


























