/*
 * Copyright 1996, 1997, 1998 Hans Reiser, see reiserfs/README for licensing and copyright details
 */
#ifdef __KERNEL__

#include <stdarg.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/reiserfs_fs.h>
#include <linux/string.h>

#else

#include "nokernel.h"
#include <stdarg.h>
/*
#define is_buffer_unwritten(bh) 0
#define is_buffer_preserved(bh) 0
#define is_buffer_suspected_recipient(bh) 0
*/

#endif


static char error_buf[1024];
static char fmt_buf[1024];

__u32 offset (struct key * key)
{
  return key->k_offset & 0x7fffffff;
}


static void sprintf_key (char * buf, struct key * key)
{
  sprintf (buf, "[%d:%d %d %d %d:%d]",
	   key->group_locale.major, key->group_locale.minor, key->name_hash, key->file_id, 
	   key->k_offset & 0x80000000, offset (key));
}


static void sprintf_item_head (char * buf, struct item_head * ih)
{
  sprintf_key (buf, &(ih->ih_key));
  sprintf (buf + strlen (buf), ", free space %d, type 0x%x, length %d, location %d",
	   ih->ih_free_space, ih->ih_type, ih->ih_item_len, ih->ih_item_location);
}


static char * is_there_reiserfs_struct (char * fmt, int * what, int * skip)
{
  char * k = fmt;

  *skip = 0;
  
  while (1) {
    k = strstr (k, "%");
    if (!k)
      break;
    if (k && (k[1] == 'k' || k[1] == 'h')) {
      *what = k[1];
      break;
    }
    (*skip) ++;
    k ++;
  }
  return k;
}


#define do_reiserfs_warning \
{\
  char * fmt1 = fmt_buf;\
  va_list args;\
  int i, j;\
  char * k;\
  char * p = error_buf;\
  int what, skip;\
\
  strcpy (fmt1, fmt);\
  va_start(args, fmt);\
\
  while (1) {\
    k = is_there_reiserfs_struct (fmt1, &what, &skip);\
    if (k != 0) {\
      *k = 0;\
      p += vsprintf (p, fmt1, args);\
\
      for (i = 0; i < skip; i ++)\
	j = va_arg (args, int);\
\
      switch (what) {\
      case 'k':\
	sprintf_key (p, va_arg(args, struct key *));\
	break;\
      case 'h':\
	sprintf_item_head (p, va_arg(args, struct item_head *));\
	break;\
      }\
      p += strlen (p);\
      fmt1 = k + 2;\
    } else {\
      i = vsprintf (p, fmt1, args);\
      break;\
    }\
  }\
\
  va_end(args);\
}


/* in addition to usual conversion specifiers this accepts reiserfs specific conversion specifiers:
   %k to print key, %h to print item_head */
void reiserfs_warning (const char * fmt, ...)
{
  do_reiserfs_warning;
  printk ("%s", error_buf);
}


/* The format:

           maintainer-errorid: [function-name:] message

    where errorid is unique to the maintainer and function-name is
    optional, is recommended, so that anyone can easily find the bug
    with a simple grep for the short to type string
    maintainer-errorid.  Don't bother with reusing errorids, there are
    lots of numbers out there.

    Example: 
    
    reiserfs_panic(
	p_sb, "reiser-29: reiserfs_new_blocknrs: "
	"one of search_start or rn(%d) is equal to MAX_B_NUM,"
	"which means that we are optimizing location based on the bogus location of a temp buffer (%p).", 
	rn, bh
    );

    Regular panic()s sometimes clear the screen before the message can
    be read, thus the need for the while loop.  

    Numbering scheme for panic used by Vladimir and Anatoly( Hans completely ignores this scheme, and considers it
    pointless complexity):

    panics in reiserfs_fs.h have numbers from 1000 to 1999
    super.c				      2000 to 2999
    preserve.c				      3000 to 3999
    bitmap.c				      4000 to 4999
    stree.c				      5000 to 5999
    prints.c				      6000 to 6999
    namei.c                                   7000 to 7999
    fix_nodes.c                               8000 to 8999
    dir.c                                     9000 to 9999
    lbalance.c				      10000 to 10999
    ibalance.c		                      11000 to 11999
    do_balan.c		                      12000 to 12999
    inode.c			              13000 to 13999
    file.c			              14000 to 14999
    objectid.c                                15000 - 15999
    buffer.c                                  16000 - 16999
    symlink.c                                 17000 - 17999

   .  */




void reiserfs_panic (struct super_block * sb, const char * fmt, ...)
{
  do_reiserfs_warning;
  printk ("%s", error_buf);

#ifdef __KERNEL__
  if (sb && !(sb->s_flags & MS_RDONLY)) {
    sb->u.reiserfs_sb.s_mount_state |= REISERFS_ERROR_FS;
    sb->u.reiserfs_sb.s_rs->s_state = REISERFS_ERROR_FS;
    
    mark_buffer_dirty(sb->u.reiserfs_sb.s_sbh, 1);
    sb->s_dirt = 1;
  }


  /* this is to prevent panic from syncing this filesystem */
  if (sb && sb->s_lock)
    sb->s_lock=0;
  if (sb)
    sb->s_flags |= MS_RDONLY;

  panic ("reiserfs_panic (device %s): %s\n",
	 sb ? kdevname(sb->s_dev) : "sb == 0", error_buf);
#else
  exit (0);
#endif
}




static char * vi_type (struct virtual_item * vi)
{
  static char *types[]={"entry", "direct", "indirect", "stat data"};

  if (vi->vi_type & (1 << IH_Stat_data))
    return types[3];
  if (vi->vi_type & (1 << IH_Indirect))
    return types[2];
  if (vi->vi_type & (1 << IH_Direct))
    return types[1];
  if (vi->vi_type & (1 << IH_Dir_entry))
    return types[0];

  reiserfs_panic (0, "vi_type: 6000: unknown type (0x%x)", vi->vi_type);
  return NULL;
}

void print_virtual_node (struct virtual_node * vn)
{
  int i, j;
  

  printk ("VIRTUAL NODE CONTAINS %d items, has size %d,%s,%s, ITEM_POS=%d POS_IN_ITEM=%d MODE=\'%c\'\n",
	  vn->vn_nr_item, vn->vn_size,
	  (vn->vn_vi[0].vi_type & (1 << VI_Leftmergeable) )? "left mergeable" : "", 
	  (vn->vn_vi[vn->vn_nr_item - 1].vi_type & (1 << VI_Rightmergeable) ) ? "right mergeable" : "",
	  vn->vn_affected_item_num, vn->vn_pos_in_item, vn->vn_mode);


  for (i = 0; i < vn->vn_nr_item; i ++)
    {
      printk ("%s %d %d", vi_type (&vn->vn_vi[i]), i, vn->vn_vi[i].vi_item_len);
      if (vn->vn_vi[i].vi_entry_sizes)
	{
	  printk ("It is directory with %d entries: ", vn->vn_vi[i].vi_entry_count);
	  for (j = 0; j < vn->vn_vi[i].vi_entry_count; j ++)
	    printk ("%d ", vn->vn_vi[i].vi_entry_sizes[j]);
	}
      printk ("\n");
    }
}


void print_path (struct tree_balance * tb, struct path * path)
{
  int h = 0;
  
  if (tb) {
    while (tb->insert_size[h]) {
      printk ("block %lu (level=%d), position %d\n", PATH_H_PBUFFER (path, h) ? PATH_H_PBUFFER (path, h)->b_blocknr : 0,
	      PATH_H_PBUFFER (path, h) ? B_BLK_HEAD (PATH_H_PBUFFER (path, h))->blk_level : 0,
	      PATH_H_POSITION (path, h));
      h ++;
    }
  } else {
    int offset = path->path_length;
    struct buffer_head * bh;
    printk ("Offset    Bh     (b_blocknr, b_count) Position Nr_item\n");
    while ( offset > ILLEGAL_PATH_ELEMENT_OFFSET ) {
      bh = PATH_OFFSET_PBUFFER (path, offset);
      printk ("%6d %10p (%9lu, %7d) %8d %7d\n", offset, 
	      bh, bh ? bh->b_blocknr : 0, bh ? bh->b_count : 0,
	      PATH_OFFSET_POSITION (path, offset), bh ? B_NR_ITEMS (bh) : -1);

      offset --;
    }
  }

#if 0
  printk ("#####################\n");
  printk ("Offset    Bh     (b_blocknr, b_count) Position Nr_item\n");
  while ( offset > ILLEGAL_PATH_ELEMENT_OFFSET ) {
    bh = PATH_OFFSET_PBUFFER (path, offset);
    printk ("%6d %10p (%9lu, %7d) %8d %7d\n", offset, 
	    bh, bh ? bh->b_blocknr : 0, bh ? bh->b_count : 0,
	    PATH_OFFSET_POSITION (path, offset), bh ? B_NR_ITEMS (bh) : -1);

    print_buffer_head(bh,"print_path");
    offset --;
  }
  printk ("#####################\n");
#endif
}



void print_block_head (struct buffer_head * bh, char * mes)
{ 

  printk ("%s: blk_level=%d, blk_nr_items=%d, blk_free_space=%d\n",
          mes, B_BLK_HEAD (bh)->blk_level, B_NR_ITEMS (bh), B_BLK_HEAD (bh)->blk_free_space);
}


void print_buffer_head(
	struct buffer_head 	* p_s_bh,
	char 			* p_c_mes
	) {
	printk ("print_buffer_head=> %s: %p\n",  p_c_mes, p_s_bh );
	if (p_s_bh == 0)
	  return;

	printk("b_data=%p, b_blocknr = %lu, ",p_s_bh->b_data, p_s_bh->b_blocknr);
	printk ("b_dev=0x%x, ", p_s_bh->b_dev);
#if 0
	kdev_t b_dev;			/* device (B_FREE = free) */
	kdev_t b_rdev;			/* Real device */
	unsigned long b_rsector;	/* Real buffer location on disk */
	struct buffer_head * b_next;	/* Hash queue list */
	struct buffer_head * b_this_page;	/* circular list of buffers in one page */
#endif

	/* Second cache line: */
	printk("b_state = 0%lo(%s %s %s %s %s %s), ", p_s_bh->b_state, 
	       buffer_uptodate (p_s_bh) ? "UPTODATE" : "NOTUPTODATE",
	       buffer_dirty (p_s_bh) ? "DIRTY" : "CLEAN",
	       buffer_locked (p_s_bh) ? "LOCKED" : "UNLOCKED",
	       is_buffer_unwritten (p_s_bh) ? "UNWR" : "! UNWR",
	       is_buffer_preserved (p_s_bh) ? "PRES" : "! PRES",
	       is_buffer_suspected_recipient (p_s_bh) ? "SUSP RECIP" : "! SUSP RECIP"
	       );

#if 0
	struct buffer_head * b_next_free;
#endif

	printk("b_count = %d, ",p_s_bh->b_count);
	printk("b_size = %lu\n",p_s_bh->b_size);


#if 0
	/* Non-performance-critical data follows. */
	char * b_data;			/* pointer to data block (1024 bytes) */
	unsigned int b_list;		/* List that this buffer appears */
	unsigned long b_flushtime;      /* Time when this (dirty) buffer
					 * should be written */
	unsigned long b_lru_time;       /* Time when this buffer was 
					 * last used. */
	struct wait_queue * b_wait;
	struct buffer_head ** b_pprev;		/* doubly linked list of hash-queue */
	struct buffer_head * b_prev_free;	/* doubly linked list of buffers */
	struct buffer_head * b_reqnext;		/* request queue */
#endif

}


/* this prints internal nodes (4 keys/items in line) (dc_number,
   dc_size)[k_dirid, k_objectid, k_offset, k_uniqueness](dc_number,
   dc_size)...*/
static void print_internal (struct buffer_head * bh, int first, int last)
{
  struct key * key;
  struct disk_child * dc;
  int i;
  int from, to;

  if (first == -1) {
    from = 0;
    to = B_NR_ITEMS (bh);
  } else {
    from = first;
    to = last < B_NR_ITEMS (bh) ? last : B_NR_ITEMS (bh);
  }

  printk ("INTERNAL NODE (%ld) contains %d keys, FREE SPACE=%d\n", 
	  bh->b_blocknr, B_NR_ITEMS (bh), B_BLK_HEAD (bh)->blk_free_space);

  dc = B_N_CHILD (bh, from);
  printk ("(%d: %lu, %u)", from, dc->dc_block_number, dc->dc_size);

  for (i = from, key = B_N_PDELIM_KEY (bh, from), dc ++; i < to; i ++, key ++, dc ++) {
    reiserfs_warning ("[%d: %k]", i, key);
    printk ("(%d: %lu, %u)", i + 1, dc->dc_block_number, dc->dc_size);
    if (i && i % 4 == 0)
      printk ("\n");
  }
  printk ("\n");
}




void print_dc (struct disk_child * dc)
{
  printk ("[dc_number=%lu, dc_size=%u]\n", dc->dc_block_number, dc->dc_size);
}




void print_bi (struct buffer_info * bi, char * mes)
{
  printk ("%s: bh->b_blocknr=%lu, bh->b_item_order=%d, bh->b_parent->b_blocknr=%lu\n", 
	  mes ? mes : "print_bi", bi->bi_bh->b_blocknr, bi->bi_position, bi->bi_parent ? bi->bi_parent->b_blocknr : 0);
}



/* these are possible types of items: 
   "sd" - fair stat data,
   "in" - indirect item,
   "di" - direct item,
   "es" - entry with stat data,
   e" - entry without stat data
*/

static char * item_type (struct item_head * ih)
{
  static char * codes[] = {"sd", "in", "di", "es", "e ", "??"};

  if (I_IS_STAT_DATA_ITEM(ih)) {
    if (!is_key_of_stat_data (&(ih->ih_key)))
      reiserfs_warning ("key type does npt match to the item type (stat data)\n");
    return codes[0];
  }
  if (I_IS_INDIRECT_ITEM(ih)) {
    if (!is_key_of_indirect_item (&(ih->ih_key)))
      reiserfs_warning ("key type does not match to the item type (indirect)\n");
    return codes[1];
  }
  if (I_IS_DIRECT_ITEM(ih)) {
    if (!is_key_of_direct_item (&(ih->ih_key)))
      reiserfs_warning ("key type does not match to the item type (direct)\n");
    return codes[2];
  }
  if (I_IS_DIRECTORY_ITEM(ih)) {
    if (!is_key_of_dir_entry (&(ih->ih_key)))
      reiserfs_warning ("key type does not match to the item type (dir entry)\n");
    if (is_entry_with_stat_data (ih))
      return codes[3];
    else
      return codes[4];
  }
  return codes[5];
}


void print_indirect_item (struct buffer_head * bh, int item_num)
{
  struct item_head * ih;
  int j;

  ih = B_N_PITEM_HEAD (bh, item_num);

  if (ih->ih_item_len % UNFM_P_SIZE)
    printk ("print_indirect_item: invalid item len");
  
  printk ("%d pointers\n[ ", ih->ih_item_len / UNFM_P_SIZE);
  for (j = 0; j < (int)(ih->ih_item_len / UNFM_P_SIZE); j ++) {
    printk ("%lu ", ((unsigned long *)B_I_PITEM (bh, ih))[j]);
  }
  printk ("]\n");
}


void print_dir_entry_item (struct buffer_head * bh, struct item_head * ih)
{
  int  i;
  struct stat_data * sd;

  printk ("|%s \"", is_entry_pointing_to_dir (ih) ? "<DIR>" : "     ");
  for (i = 0; i < B_I_NAMELEN (bh, ih); i ++)
    printk ("%c", B_I_NAME (bh, ih)[i]);
  if (is_entry_pointing_to_dir (ih))
    reiserfs_warning ("\"\t%d", ih->ih_key.file_id);
  else
    reiserfs_warning ("\"\t%k", B_I_OBJECT_LOCALE_FROM_ENTRY (bh, ih));
  if (is_entry_with_stat_data (ih)) {
    sd = B_I_STAT_DATA (bh, ih);
    printk ("\t0%-6o %6u %2u (first direct byte %d)", sd->sd_mode, sd->sd_size, sd->sd_nlink, sd->u.first_direct_byte);
  }
  printk ("\n");
}


static int is_symlink = 0;
static void print_leaf (struct buffer_head * bh, int print_mode, int first, int last)
{
  struct block_head * blkh;
  struct item_head * ih;
  int i;
  int from, to;

  blkh = B_BLK_HEAD (bh);
  ih = B_N_PITEM_HEAD (bh,0);
           /*123456789012345678901234567890123456789012345678901234567890123456789012*/
  printk ("\n========================================================================\n");
  reiserfs_warning ("LEAF NODE (%ld) contains NR_ITEM=%d, FREE SPACE=%d, RIGHT_DELIM_KEY: %k\n", 
		    bh->b_blocknr, blkh->blk_nr_item, blkh->blk_free_space, B_PRIGHT_DELIM_KEY (bh));

  if (!(print_mode & PRINT_LEAF_ITEMS)) {
    reiserfs_warning ("FIRST ITEM_KEY: %k\nLAST ITEM KEY: %k\n", &(ih->ih_key), &(ih[blkh->blk_nr_item - 1].ih_key));
    return;
  }

  if (first < 0 || first > blkh->blk_nr_item - 1) 
    from = 0;
  else 
    from = first;

  if (last < 0 || last > blkh->blk_nr_item)
    to = blkh->blk_nr_item;
  else
    to = last;

         /*123456789012345678901234567890123456789012345678901234567890123456789012*/
  printk ("------------------------------------------------------------------------\n");
  printk ("|##|  |       key                 |  item   |      stat data           |\n");
  printk ("|  |  |-----------------------------------------------------------------\n");
  printk ("|  |  |locale hash    file_id offset      | loc len | mode  size  nl           |\n");

  for (i = from; i < to; i++) {
    printk ("------------------------------------------------------------------------\n");
    reiserfs_warning ("|%2d|%s| %h |", i, item_type(ih+i), &(ih[i]));

    if (I_IS_STAT_DATA_ITEM(ih+i)) {
      struct stat_data * sd = B_I_STAT_DATA (bh,ih+i);

      printk ("0%-6o %6u %2u\n", sd->sd_mode, sd->sd_size, sd->sd_nlink);
      is_symlink = (S_ISLNK(sd->sd_mode)) ? 1 : 0;
      continue;
    }
    printk ("\n");
    if (I_IS_DIRECTORY_ITEM(ih+i) && print_mode & PRINT_DIRECTORY_ITEMS) {
      print_dir_entry_item (bh, ih+i);
      continue;
    }

    if (I_IS_INDIRECT_ITEM(ih+i)) {
      print_indirect_item (bh, i);
      continue;
    }

    if (I_IS_DIRECT_ITEM(ih+i)) {
      int j = 0;
      if (is_symlink || print_mode & PRINT_DIRECT_ITEMS) {
	printk ("\"");
	while (j < ih[i].ih_item_len)
	  printk ("%c", B_I_PITEM(bh,ih+i)[j++]);
	printk ("\"\n");
      }
      continue;
    }
  }
  printk ("========================================================================\n");
  
}


void print_block (struct buffer_head * bh, int print_mode, int first, int last)
{
  if ( ! bh ) {
    printk("print_block: buffer is NULL\n");
    return;
  }

  /*print_buffer_head(bh,"print_block");*/

  if (B_IS_ITEMS_LEVEL(bh))
    print_leaf (bh, print_mode, first, last);
  else if (B_IS_KEYS_LEVEL(bh))
    print_internal (bh, first, last);
  else
    printk ("block %ld has invalid type (blk_tree_level == %d)\n", 
	    bh->b_blocknr, B_BLK_HEAD(bh)->blk_level);
}


void print_tb (int mode, int item_pos, int pos_in_item, struct tree_balance * tb, char * mes)
{
  int h = 0;
  int i;
  struct buffer_head * tbSh, * tbFh;


  if (!tb)
    return;

  printk ("\n********************** PRINT_TB for %s *******************\n", mes);
  printk ("MODE=%c, ITEM_POS=%d POS_IN_ITEM=%d\n", mode, item_pos, pos_in_item);
  printk ("*********************************************************************\n");

  printk ("* h *    S    *    L    *    R    *   F   *   FL  *   FR  *  CFL  *  CFR  *\n");
/*
01234567890123456789012345678901234567890123456789012345678901234567890123456789
       1        2         3         4         5         6         7         8
  printk ("*********************************************************************\n");
*/
  
  
  for (h = 0; h < sizeof(tb->insert_size) / sizeof (tb->insert_size[0]); h ++) {
    if (PATH_H_PATH_OFFSET (tb->tb_path, h) <= tb->tb_path->path_length && 
	PATH_H_PATH_OFFSET (tb->tb_path, h) > ILLEGAL_PATH_ELEMENT_OFFSET) {
      tbSh = PATH_H_PBUFFER (tb->tb_path, h);
      tbFh = PATH_H_PPARENT (tb->tb_path, h);
    } else {
      /*      printk ("print_tb: h=%d, PATH_H_PATH_OFFSET=%d, path_length=%d\n", 
	      h, PATH_H_PATH_OFFSET (tb->tb_path, h), tb->tb_path->path_length);*/
      tbSh = 0;
      tbFh = 0;
    }
    printk ("* %d * %3ld(%2d) * %3ld(%2d) * %3ld(%2d) * %5ld * %5ld * %5ld * %5ld * %5ld *\n",
	    h, 
	    (tbSh) ? (tbSh->b_blocknr):(-1),
	    (tbSh) ? tbSh->b_count : -1,
	    (tb->L[h]) ? (tb->L[h]->b_blocknr):(-1),
	    (tb->L[h]) ? tb->L[h]->b_count : -1,
	    (tb->R[h]) ? (tb->R[h]->b_blocknr):(-1),
	    (tb->R[h]) ? tb->R[h]->b_count : -1,
	    (tbFh) ? (tbFh->b_blocknr):(-1),
	    (tb->FL[h]) ? (tb->FL[h]->b_blocknr):(-1),
	    (tb->FR[h]) ? (tb->FR[h]->b_blocknr):(-1),
	    (tb->CFL[h]) ? (tb->CFL[h]->b_blocknr):(-1),
	    (tb->CFR[h]) ? (tb->CFR[h]->b_blocknr):(-1));
  }

  printk ("*********************************************************************\n");


  /* print balance parameters for leaf level */
  h = 0;
  printk ("* h * size * ln * lb * rn * rb * blkn * s0 * s1 * s1b * s2 * s2b * curb * lk * rk *\n");
  printk ("* %d * %4d * %2d * %2d * %2d * %2d * %4d * %2d * %2d * %3d * %2d * %3d * %4d * %2d * %2d *\n",
	  h, tb->insert_size[h], tb->lnum[h], tb->lbytes, tb->rnum[h],tb->rbytes, tb->blknum[h], 
	  tb->s0num, tb->s1num,tb->s1bytes,  tb->s2num, tb->s2bytes, tb->cur_blknum, tb->lkey[h], tb->rkey[h]);


/* this prints balance parameters for non-leaf levels */
  do {
    h++;
    printk ("* %d * %4d * %2d *    * %2d *    * %2d *\n",
    h, tb->insert_size[h], tb->lnum[h], tb->rnum[h], tb->blknum[h]);
  } while (tb->insert_size[h]);

  printk ("*********************************************************************\n");


  /* print FEB list (list of buffers in form (bh (b_blocknr, b_count), that will be used for new nodes) */
  h = 0;
  for (i = 0; i < sizeof (tb->FEB) / sizeof (tb->FEB[0]); i ++)
    printk ("%s%p (%lu %d)", i == 0 ? "FEB list: " : ", ", tb->FEB[i], tb->FEB[i] ? tb->FEB[i]->b_blocknr : 0,
	    tb->FEB[i] ? tb->FEB[i]->b_count : 0);
  printk ("\n");

  printk ("********************** END OF PRINT_TB *******************\n\n");

}


#ifndef __KERNEL__

/* 
 * read bitmap blocks and print them
 */
void print_bmap (int bmapnr, char ** bmap, int blocksize, char * mes)
{
  int i, j;
  int bits = blocksize * 8;

  if (mes)
    printf ("%s\n", mes);
  for (i = 0; i < bmapnr; i ++) {
    if (test_bit (0, bmap[i])) {
      /* first block addressed by this bitmap block is used */
      printf ("#%d Busy (%d-", i, i * bits);
      for (j = 1; j < bits; j ++) {
	while (test_bit (j, bmap[i])) {
	  if (j == bits - 1) {
	    printf ("%d)\n", j + i * bits);
	    goto end;
	  }
	  j++;
	}
	printf ("%d) Free(%d-", j - 1 + i * bits, j + i * bits);

	while (!test_bit (j, bmap[i])) {
	  if (j == bits - 1) {
	    printf ("%d)\n", j + i * bits);
	    goto end;
	  }
	  j++;
	}
	printf ("%d) Busy(%d-", j - 1 + i * bits, j + i * bits);
	
      end:
      }
    } else {
      /* first block addressed by this bitmap is free */
      printf ("#%d Free (%d-", i, i * bits);
      for (j = 1; j < bits; j ++) {
	while (!test_bit (j, bmap[i])) {
	  if (j == bits - 1) {
	    printf ("%d)\n", j + i * bits);
	    goto end2;
	  }
	  j++;
	}
	printf ("%d) Busy(%d-", j - 1 + i * bits, j + i * bits);

	while (test_bit (j, bmap[i])) {
	  if (j == bits - 1) {
	    printf ("%d)\n", j + i * bits);
	    goto end2;
	  }
	  j++;
	}
	printf ("%d) Busy(%d-", j - 1 + i * bits, j + i * bits);
	
      end2:
      }
    }
  }
}


void print_super_block (struct super_block * s, char * mes)
{
  struct reiserfs_super_block * rs;

  rs = SB_DISK_SUPER_BLOCK (s);
  printf ("========= %s =============\n", mes);
  printf ("-------------------------------------------------------------------------------------------------------------------\n");
  printf ("| block count | free block | used block | blocksize | tree height |  state  |   magic   | root block | bmap count |\n");
  printf ("|             |   count    |  count     |           |             |         |           |            |            |\n");
  printf ("|-----------------------------------------------------------------------------------------------------------------|\n");
  printf ("| %11d | %10d | %10d | %9d | %11d | %7s | %9s | %10d | %10d |\n", 
	  rs->s_block_count, rs->s_free_blocks, rs->s_block_count - rs->s_free_blocks, rs->s_blocksize, rs->s_tree_height, 
	  (rs->s_state == REISERFS_VALID_FS) ? "VALID" : "ERROR", rs->s_magic, rs->s_root_block, rs->s_bmap_nr);
  printf ("|-----------------------------------------------------------------------------------------------------------------|\n\n");
}


void print_objectid_map (struct super_block * s)
{
  int i;
  struct reiserfs_super_block * rs;
  unsigned long * omap;

  rs = SB_DISK_SUPER_BLOCK (s);
  omap = (unsigned long *)(rs + 1);
  printk ("Map of objectids\n");
      
  for (i = 0; i < rs->s_oid_cursize; i ++) {
    if (i % 2 == 0)
      printk ("busy(%lu-%lu) ", omap[i], omap[i+1] - 1); 
    else
      printk ("free(%lu-%lu) ", 
	      omap[i], ((i+1) == rs->s_oid_cursize) ? -1 : omap[i+1] - 1);
    }
  printk ("\n");
  
  printk ("Object id array has size %d (max %d):", rs->s_oid_cursize, 
	  rs->s_oid_maxsize);
  
  for (i = 0; i < rs->s_oid_cursize; i ++)
    printk ("%lu ", omap[i]); 
  printk ("\n");

}

#endif	/* #ifndef __KERNEL__ */
