/*
 * Copyright 1996, 1997, 1998 Hans Reiser, see reiserfs/README for licensing and copyright details
 */
#ifdef __KERNEL__

#include <linux/module.h>

#include <linux/sched.h>
#include <asm/uaccess.h>
#include <linux/reiserfs_fs.h>
#include <linux/locks.h>
#include <linux/init.h>

#else

#include "nokernel.h"

#endif

void reiserfs_write_super (struct super_block * s)
{
  struct reiserfs_super_block * rs;

  if (!(s->s_flags & MS_RDONLY)) {
    rs = SB_DISK_SUPER_BLOCK (s);
    /*
     * if reiserfs was mounted with read-write permissions make file
     * system state not valid so that if we crash without doing a
     * clean umount we know that we must run file system
     * checker. umount will mark it valid if it does a clean umount
     */
    if (le16_to_cpu (rs->s_state) == REISERFS_VALID_FS) {
      rs->s_state = cpu_to_le16 (REISERFS_ERROR_FS);
      mark_buffer_dirty (SB_BUFFER_WITH_SB (s), 1);
    }
  }
  s->s_dirt = 0;
}


void reiserfs_put_super (struct super_block * s)
{
  int i;
  kdev_t dev = s->s_dev;

  /* change file system state to current state if it was mounted with read-write permissions */
  if (!(s->s_flags & MS_RDONLY)) {
    SB_REISERFS_STATE (s) = le16_to_cpu (s->u.reiserfs_sb.s_mount_state);
    mark_buffer_dirty (SB_BUFFER_WITH_SB (s), 1);
  }

  if (maybe_free_preserve_list (s) == 0) {
    reiserfs_warning ("vs-2000: reiserfs_put_super: there are %ld buffers to write\n", s->u.reiserfs_sb.s_suspected_recipient_count);

    mark_suspected_recipients_dirty (dev);
    fsync_dev (dev);
    s->u.reiserfs_sb.s_suspected_recipient_count = 0;
    maybe_free_preserve_list (s);
  }

#ifdef REISERFS_CHECK
  preserve_trace_release_bitmap (s);
#endif

  for (i = 0; i < SB_BMAP_NR (s); i ++) {
    /* update cautious bitmap */
    if (memcmp (SB_AP_TRUE_BITMAP (s)[i]->b_data, SB_AP_CAUTIOUS_BITMAP (s)[i], SB_AP_TRUE_BITMAP (s)[i]->b_size)) {
      memcpy (SB_AP_CAUTIOUS_BITMAP (s)[i]->b_data, SB_AP_TRUE_BITMAP (s)[i]->b_data, SB_AP_TRUE_BITMAP (s)[i]->b_size);
      mark_buffer_dirty (SB_AP_CAUTIOUS_BITMAP (s)[i], 1);
      ll_rw_block (WRITE, 1, &SB_AP_CAUTIOUS_BITMAP (s)[i]);
    }
  }
  /* wait on write completion */
  for (i = 0; i < SB_BMAP_NR (s); i ++) {
    wait_on_buffer (SB_AP_CAUTIOUS_BITMAP (s)[i]);
    brelse (SB_AP_CAUTIOUS_BITMAP (s)[i]);
    brelse (SB_AP_TRUE_BITMAP (s)[i]);
  }

  kfree (SB_AP_TRUE_BITMAP (s));
  kfree (SB_AP_CAUTIOUS_BITMAP (s));


/*  release_log_buffer (s);*/

  brelse (SB_BUFFER_WITH_SB (s));

  /*
  printk ("reiserfs_put_super: session statistics: balances %d, fix_nodes %d, preserve list freeings %d, \
bmap with search %d, without %d, dir2ind %d, ind2dir %d\n",
	  s->u.reiserfs_sb.s_do_balance, s->u.reiserfs_sb.s_fix_nodes, s->u.reiserfs_sb.s_preserve_list_freeings,
	  s->u.reiserfs_sb.s_bmaps, s->u.reiserfs_sb.s_bmaps_without_search,
	  s->u.reiserfs_sb.s_direct2indirect, s->u.reiserfs_sb.s_indirect2direct);
  */


  s->s_dev = 0;

  invalidate_reiserfs_buffers (dev);

#ifdef REISERFS_CHECK
  reiserfs_show_buffers (dev);
#endif
  /*  unlock_super (s);*/

  MOD_DEC_USE_COUNT;
  return;
}


/* super block operations are */
static struct super_operations reiserfs_sops = 
{
  reiserfs_read_inode,
  reiserfs_write_inode,
  NULL,				/* put_inode*/
  reiserfs_delete_inode,
  reiserfs_notify_change,
  reiserfs_put_super,
  reiserfs_write_super,
  reiserfs_statfs,
  reiserfs_remount,
  NULL, 				/* clear_inode */
  NULL				/* umount_begin */
};


int reiserfs_remount (struct super_block * s, int * flags, char * data)
{
  struct reiserfs_super_block * rs;

  rs = SB_DISK_SUPER_BLOCK (s);
  if ((unsigned long)(*flags & MS_RDONLY) == (s->s_flags & MS_RDONLY))
    /* there is nothing to do to remount read-only fs as read-only fs */
    return 0;

  if (*flags & MS_RDONLY) {
    /* try to remount file system with read-only permissions */
    if (le16_to_cpu (rs->s_state) == REISERFS_VALID_FS ||
	s->u.reiserfs_sb.s_mount_state != REISERFS_VALID_FS)
      return 0;

    /* Mounting a rw partition read-only. */
    rs->s_state = cpu_to_le16 (s->u.reiserfs_sb.s_mount_state);
    mark_buffer_dirty (SB_BUFFER_WITH_SB (s), 1);
    s->s_dirt = 0;
  } else {
    /* Mount a partition which is read-only, read-write */
    s->u.reiserfs_sb.s_mount_state = le16_to_cpu (rs->s_state);
    s->s_flags &= ~MS_RDONLY;
    rs->s_state = cpu_to_le16 (REISERFS_ERROR_FS);
    mark_buffer_dirty (SB_BUFFER_WITH_SB (s), 1);
    s->s_dirt = 0;
    
    /* check state, which file system had when remounting read-write */
    if (s->u.reiserfs_sb.s_mount_state != REISERFS_VALID_FS)
      printk ("REISERFS: remounting unchecked fs, "
	      "running reiserfsck is recommended\n");
  }
  return 0;
}


struct key root_key = {{0, REISERFS_ROOT_OBJECTID}, 0, REISERFS_ROOT_OBJECTID, 0};


int read_bitmaps (struct super_block * s)
{
  int i, repeat;
  struct reiserfs_super_block * rs = SB_DISK_SUPER_BLOCK(s);
  int bmp1 = (REISERFS_FIRST_BLOCK * BLOCK_SIZE) / s->s_blocksize + 1;	/* first of bitmap blocks */

  /* read true bitmap */
  SB_AP_TRUE_BITMAP (s) = kmalloc (sizeof (struct buffer_head *) * le16_to_cpu (rs->s_bmap_nr), GFP_KERNEL);
  if (SB_AP_TRUE_BITMAP (s) == 0)
    return 1;

  memset (SB_AP_TRUE_BITMAP (s), 0, sizeof (struct buffer_head *) * le16_to_cpu (rs->s_bmap_nr));

  for (i = 0; i < le16_to_cpu (rs->s_bmap_nr); i ++) {
    SB_AP_TRUE_BITMAP (s)[i] = reiserfs_bread (s->s_dev, bmp1 + i, s->s_blocksize, &repeat);
    if (!SB_AP_TRUE_BITMAP (s)[i])
      return 1;
  }

  /* read cautious bitmap, which must be equal to the true one */
  SB_AP_CAUTIOUS_BITMAP (s) = kmalloc (sizeof (struct buffer_head *) * le16_to_cpu (rs->s_bmap_nr), GFP_KERNEL);
  if (SB_AP_CAUTIOUS_BITMAP (s) == 0)
    return 1;

  memset (SB_AP_CAUTIOUS_BITMAP (s), 0, sizeof (struct buffer_head *) * le16_to_cpu (rs->s_bmap_nr));

  for (i = 0; i < le16_to_cpu (rs->s_bmap_nr); i ++) {
    SB_AP_CAUTIOUS_BITMAP (s)[i] = reiserfs_bread (s->s_dev, bmp1 + rs->s_bmap_nr + i, s->s_blocksize, &repeat);
    if (!SB_AP_CAUTIOUS_BITMAP (s)[i])
      return 1;
    if (memcmp (SB_AP_CAUTIOUS_BITMAP (s)[i]->b_data, SB_AP_TRUE_BITMAP (s)[i]->b_data, s->s_blocksize)) {
      printk ("reiserfs: read_bitmaps: %d blocks differ in true and cautious bitmaps\n", i);
      memcpy (SB_AP_CAUTIOUS_BITMAP (s)[i]->b_data, SB_AP_TRUE_BITMAP (s)[i]->b_data, s->s_blocksize);
      mark_buffer_dirty (SB_AP_CAUTIOUS_BITMAP (s)[i], 1);
    }
  }

  return 0;
}


/* this was (ext2)parse_options */
static int parse_options (char * options, unsigned long * mount_options)
{
  char * this_char;
  
  if (!options)
    /* use default configuration: complex read, create tails, preserve on */
    return 1;
  for (this_char = strtok (options, ","); this_char != NULL; this_char = strtok (NULL, ",")) {
    if (!strcmp (this_char, "notail")) {
      set_bit (NOTAIL, mount_options);
    } else if (!strcmp (this_char, "nopreserve")) {
      set_bit (NOPRESERVE, mount_options);
    } else if (!strcmp (this_char, "genericread")) {
      set_bit (GENERICREAD, mount_options);
    } else {
      printk ("reiserfs: Unrecognized mount option %s\n", this_char);
      return 0;
    }
  }
  return 1;
}




struct super_block * reiserfs_read_super (struct super_block * s, void * data, int silent)
{
  int size;
  struct buffer_head * bh;
  struct reiserfs_super_block * rs;
  struct inode *root_inode;
  kdev_t dev = s->s_dev;
  int j;
  int repeat;
  extern int *blksize_size[];
  struct path path;

  init_path (&path);

  memset (&s->u.reiserfs_sb, 0, sizeof (struct reiserfs_sb_info));

  if (parse_options ((char *) data, &(s->u.reiserfs_sb.s_mount_opt)) == 0) {
    s->s_dev = 0;
    return NULL;
  }

  MOD_INC_USE_COUNT;
  lock_super (s);


  if (blksize_size[MAJOR(dev)] && blksize_size[MAJOR(dev)][MINOR(dev)] != 0) {
    /* as blocksize is set for partition we use it */
    size = blksize_size[MAJOR(dev)][MINOR(dev)];
  } else {
    size = BLOCK_SIZE;
    set_blocksize (s->s_dev, BLOCK_SIZE);
  }
  if (size > 8192)
    goto out_bad_sb;

  /* read block, containing reiserfs super block (it is stored at
     REISERFS_FIRST_BLOCK-th 1K block) */
  bh = bread (dev, (REISERFS_FIRST_BLOCK * BLOCK_SIZE) / size, size);
  if (!bh)
    goto out_bad_sb;

  rs = (struct reiserfs_super_block *)bh->b_data;
  if (strncmp (rs->s_magic, REISERFS_SUPER_MAGIC, strlen (REISERFS_SUPER_MAGIC)))
    goto out_no_fs;

  s->s_blocksize = le16_to_cpu (rs->s_blocksize);
  s->s_blocksize_bits = 0;
  while ((1 << s->s_blocksize_bits) != s->s_blocksize)
    s->s_blocksize_bits ++;

  if (size != rs->s_blocksize) {
    brelse (bh);
    set_blocksize (dev, s->s_blocksize);
    /*bh = bread (dev, (REISERFS_FIRST_BLOCK * BLOCK_SIZE) / s->s_blocksize, s->s_blocksize);*/
    bh = reiserfs_bread (dev, (REISERFS_FIRST_BLOCK * BLOCK_SIZE) / s->s_blocksize, s->s_blocksize, &repeat);
    if (!bh)
      goto out_bad_sb;

    rs = (struct reiserfs_super_block *)bh->b_data;
    if (strncmp (rs->s_magic, REISERFS_SUPER_MAGIC, strlen (REISERFS_SUPER_MAGIC)) ||
	le16_to_cpu (rs->s_blocksize) != s->s_blocksize)
      goto out_no_fs;    
  }

  SB_BUFFER_WITH_SB (s) = bh;
  SB_DISK_SUPER_BLOCK (s) = rs;
  s->u.reiserfs_sb.s_mount_state = le16_to_cpu (rs->s_state);

  /* reiserfs can not be mounted when it propably contains errors */
  if (le16_to_cpu (rs->s_state) != REISERFS_VALID_FS)
    goto out_check_first;

  if (read_bitmaps (s))
    goto out_no_bitmap;

/*
  if (init_log_buffer (s))
    goto out_freemap;
*/

  s->s_op = &reiserfs_sops;
   
  /* get root directory inode */
  root_inode = reiserfs_iget (s, &path, &root_key);
  if (!root_inode)
    goto out_no_root;

  s->s_root = d_alloc_root(root_inode, NULL);  
  if (!s->s_root)
    goto out_iput;

  if (!(s->s_flags & MS_RDONLY)) {
    rs->s_state = cpu_to_le16 (REISERFS_ERROR_FS);
    mark_buffer_dirty (bh, 1);
    s->s_dirt = 0;
  }

  s->u.reiserfs_sb.unpreserve = dont_preserve (s) ? 0 : unpreserve;

#ifdef REISERFS_CHECK
  preserve_trace_init_bitmap (s);
#endif

  unlock_super (s);
  return s;

out_iput:
  iput(root_inode);
  goto out_freemap;

out_no_root:
  printk ("reiserfs_read_super: get root inode failed\n");
  goto out_freemap;

out_no_bitmap:
  printk ("reiserfs_read_super: unable to read bitmap\n");

out_freemap:
  for (j = 0; j < le16_to_cpu (rs->s_bmap_nr); j ++) {
    if (SB_AP_TRUE_BITMAP (s))
      brelse (SB_AP_TRUE_BITMAP (s)[j]);
    if (SB_AP_CAUTIOUS_BITMAP (s))
      brelse (SB_AP_CAUTIOUS_BITMAP (s)[j]);
  }
  goto out_release;

  /*
out_no_mem:
  printk ("reiserfs_read_super: no enough memory for array of pointers to block bitmap buffers\n");
  goto out_release;
  */

out_check_first:
  printk ("reiserfs_read_super:  mounting unchecked fs, run reiserfsck first\n");
  goto out_release;

out_no_fs:
  if (!silent)
    printk("VFS: Can't find a reiserfs filesystem on dev %s.\n", kdevname(dev));

out_release:
  brelse(bh);
  goto out_unlock;

out_bad_sb:
  printk("reiserfs_read_super: unable to read superblock\n");

out_unlock:
  s->s_dev = 0;
  unlock_super(s);
  MOD_DEC_USE_COUNT;

  return NULL;
}


int reiserfs_statfs (struct super_block * s, struct statfs * buf, int bufsize)
{
  struct statfs tmp;
  struct reiserfs_super_block * rs = SB_DISK_SUPER_BLOCK (s);

  tmp.f_type = -1;
  tmp.f_bsize = le32_to_cpu (s->s_blocksize);
  tmp.f_blocks = le32_to_cpu (rs->s_block_count) - le16_to_cpu (rs->s_bmap_nr) - 1;
  tmp.f_bfree = le32_to_cpu (rs->s_free_blocks);
  tmp.f_bavail = tmp.f_bfree;
  /* total file nodes in file system and free file nodes in fs - these are calculated very approximately */
  tmp.f_files = (tmp.f_blocks * s->s_blocksize / (SD_SIZE + IH_SIZE));
  tmp.f_ffree = (tmp.f_bfree * s->s_blocksize / (SD_SIZE + IH_SIZE));
  tmp.f_namelen = (REISERFS_MAX_NAME_LEN (s->s_blocksize));
  return copy_to_user (buf, &tmp, bufsize) ? -EFAULT : 0;
}

#ifdef __KERNEL__

static struct file_system_type reiserfs_fs_type = {
  "reiserfs", FS_REQUIRES_DEV, reiserfs_read_super, NULL
};


__initfunc(int init_reiserfs_fs(void))
{
        return register_filesystem(&reiserfs_fs_type);
}

#endif

#ifdef MODULE
EXPORT_NO_SYMBOLS;

int init_module(void)
{
	return init_reiserfs_fs();
}

void cleanup_module(void)
{
        unregister_filesystem(&reiserfs_fs_type);
}

#endif






























