/*
 * Copyright 1996, 1997, 1998 Hans Reiser, see reiserfs/README for licensing and copyright details
 */
#ifndef _LINUX_REISER_FS_H
#define _LINUX_REISER_FS_H


#include <linux/types.h>
#ifdef __KERNEL__
#include <linux/malloc.h>
#endif

/*
 *  include/linux/reiser_fs.h
 *
 *  Reiser File System constants and structures
 *
 */

/* in reading the #defines, it may help to understand that they employ
   the following abbreviations:
   
   B = Buffer
   I = Item header
   H = Height within the tree (should be changed to LEV)
   N = Number of the item in the node
   DNM = DyNaMic data
   STAT = stat data
   DEH = Directory Entry Header
   EC = Entry Count
   E = Entry number
   UL = Unsigned Long
   BLKH = BLocK Header
   UNFM = UNForMatted node
   DC = Disk Child
   P = Path

   These #defines are named by concatenating these abbreviations, where
   first comes the arguments, and last comes the return value, of the
   macro.
   
   */

#define REISERFS_CHECK


/* NEW_GET_NEW_BUFFER will try to allocate new blocks better */
/*#define NEW_GET_NEW_BUFFER*/
#define OLD_GET_NEW_BUFFER


/* if this is undefined, all inode changes get into stat data immediately, if it can be found in RAM */
#define DIRTY_LATER

/* these are used by reiserfs_file_read */
#define REISERFS_NBUF 32

/* these say to reiserfs_file_read about desired kind of read ahead */
#define REISERFS_OBJECT_READ_AHEAD

#define PACKING_LOCALITY_READ_AHEAD

/*#define READ_LOCK_REISERFS*/


/* if you do not plan to have more, than one mounted reiserfs, turning
   this on will add more check over preserve list */
/*#define ONE_REISERFS*/


/* obsolete defines */
/*#define REISERFS_CHECK_ONE_PROCESS*/
/*#define REISERFS_INFO*/

/*#endif*/	/* __KERNEL__ */




/*
 * Disk Data Structures
 */

/***************************************************************************/
/*                             SUPER BLOCK                                 */
/***************************************************************************/

/*
 * Structure of super block on disk, a version of which in RAM is often accessed as s->u.reiserfs_sb.s_rs
 * the version in RAM is part of a larger structure containing fields never written to disk.
 */
#define REISERFS_SUPER_MAGIC "ReIsErFs"

/* reiserfs super block is stored in 8th 1K block of device __always__ */
#define REISERFS_FIRST_BLOCK 8

#define READ_BLOCKS  1
#define DONT_READ_BLOCKS 2

#define CARRY_ON          	0
#define SCHEDULE_OCCURRED  	1
#define PATH_INCORRECT    	2

#define NO_DISK_SPACE        (-1)
#define NO_BALANCING_NEEDED  (-2)


struct buffer_and_id {
  struct buffer_head  * bi_buf;
  unsigned long  bi_id;
};


struct unfm_nodeinfo {
  unsigned long	 unfm_nodenum;
  unsigned short unfm_freespace;
};


/* the defines below say, that if file size is >=
   DIRECT_TAIL_SUPPRESSION_SIZE * blocksize, then if tail is longer
   than MAX_BYTES_SUPPRESS_DIRECT_TAIL, it will be stored in
   unformatted node */
#define DIRECT_TAIL_SUPPRESSION_SIZE      1024
#define MAX_BYTES_SUPPRESS_DIRECT_TAIL    1024

/* Check whether byte is placed in a direct item. */
#define INODE_OFFSET_IN_DIRECT(p_s_inode, n_offset) \
( (n_offset) >= (p_s_inode)->u.reiserfs_i.i_first_direct_byte )

  
/* this is used for i_first_direct_byte field of inode */
#define NO_BYTES_IN_DIRECT_ITEM MAX_KEY_OFFSET
  
/* We store tail in unformatted node if it is too big to fit into a
   formatted node or if DIRECT_TAIL_SUPPRESSION_SIZE,
   MAX_BYTES_SUPPRESS_DIRECT_TAIL and file size say that. */
/* #define STORE_TAIL_IN_UNFM(n_file_size,n_tail_size,n_block_size) \ */
/* ( ((n_tail_size) > MAX_DIRECT_ITEM_LEN(n_block_size)) || \ */
/*   ( ( (n_file_size) >= (n_block_size) * DIRECT_TAIL_SUPPRESSION_SIZE ) && \ */
/*    ( (n_tail_size) >= MAX_BYTES_SUPPRESS_DIRECT_TAIL ) ) ) */

  /* This is an aggressive tail suppression policy, I am hoping it
     improves our benchmarks. The principle behind it is that
     percentage space saving is what matters, not absolute space
     saving.  This is non-intuitive, but it helps to understand it if
     you consider that the cost to access 4 blocks is not much more
     than the cost to access 1 block, if you have to do a seek and
     rotate.  A tail risks a non-linear disk access that is
     significant as a percentage of total time cost for a 4 block file
     and saves an amount of space that is less significant as a
     percentage of space, or so goes the hypothesis.  -Hans */
#define STORE_TAIL_IN_UNFM(n_file_size,n_tail_size,n_block_size) \
\
( ((n_tail_size) > MAX_DIRECT_ITEM_LEN(n_block_size)) || \
  ( (n_file_size) >= (n_block_size) * 4 ) || \
   ( ( (n_file_size) >= (n_block_size) * 3 ) && \
   ( (n_tail_size) >=   (MAX_DIRECT_ITEM_LEN(n_block_size))/4) ) || \
   ( ( (n_file_size) >= (n_block_size) * 2 ) && \
   ( (n_tail_size) >=   (MAX_DIRECT_ITEM_LEN(n_block_size))/2) ) || \
   ( ( (n_file_size) >= (n_block_size) ) && \
   ( (n_tail_size) >=   (MAX_DIRECT_ITEM_LEN(n_block_size) * 3)/4) ) )

  
/*
 * values for s_state field
 */
#define REISERFS_VALID_FS    1
#define REISERFS_ERROR_FS    2

  

/***************************************************************************
 *                       KEY & ITEM HEAD                                   *
 ***************************************************************************/
#if 0

struct file_part_id {
  __u32 file_id;  /* 0 for directory stat data. File objectid for file items.  We don't set this
		     equal to the hash of the first name the file is created with because the use of
		     hard links makes that complex when that first name is deleted and reused. */

  __u64 offset;  /* Parts of a file are ordered by offset within the file. */
};

struct name_id {
  __u32 hash_gen; /* We hash the filename so that we can have a 32 bit version of the name to save
                     space on keys and to make cookies in NFS V2 work.  When Linux goes to V3 we can
                     make this 64 bits.  This contains 24 bits of hash (see teahash.c) plus 8 bits
                     of generation number.  The generation number distinguishes identical hashes.
                     Note that we assume that the lexicographic order of the last component of a
                     filename conveys nothing about the locality of reference of the file.  To be
                     precise, our locality of reference for names is by default based solely on the
                     pack_locale which is based on the object id of the parent directory.  The
                     non-default is yet to be implemented but should be fun.... */

  __u32 file_id; /* The file_id distinguishes the key of a name from the keys of identical names
		    that have been removed but are retained to keep hard links happy.  Also, from
		    the file_id plus the pack_locale plus the offset one can compute the the key of
		    any file part, and knowing that will allow us to someday save 4 bytes from our
		    directory entries when we get a chance to re-code.  */

  __u32 offset;
};

struct key {
  struct {
    unsigned int major_pack_locale: 30; /* this groups all items of all files belonging to the same
					   directory and items-entries of the directory itself
					   together. It equals to objectid of the directory.  Note
					   that file stat datas are stored in dir entries.  */
    unsigned int minor_pack_locale: 2;  /* This groups dir items together by value 0.  indirect
					   items - by 1, direct items - by 2 */
  } pack_locale;

  union {
    struct file_part_id file;   /* keys for file parts contain this */
    struct name_id name;        /* keys for directory entries contain this */
  } u;
};

#endif /* 0 */


struct locale {
  __u32 minor: 2;
  __u32 major: 30;
};


struct key {
  struct locale group_locale;	/* this groups all items of all files belonging to the same
				   directory and items-entries of the directory itself together. It
				   equals to objectid of the directory.  Note that file stat data
				   are stored in dir entries. */
  __u32 name_hash;	/* We hash the filename so that we can have a 32 bit version of the name to
			   save space on keys and to make cookies in NFS V2 work.  When Linux goes
			   to V3 we can make this 64 bits. This contains 31 bits of hash (see
			   teahash.c). The highest bit is 0. Note that we assume that the
			   lexicographic order of the last component of a filename conveys nothing
			   about the locality of reference of the file.  To be precise, our locality
			   of reference for names is by default based solely on the group_locale
			   which is based on the object id of the parent directory.  The non-default
			   is yet to be implemented but should be fun.... */
  __u32 file_id;
  __u32 k_offset;
};

#define KEY_SIZE (sizeof(struct key))

/* stat data and dir entry items have k_offset field 0 */
#define SD_OFFSET 0



/* Everything in the filesystem is stored as a set of items.  The item head contains the key of the
   item, its free space (for indirect items) and specifies the location of the item itself within
   the block.  */
struct item_head
{
  struct key ih_key;		/* Everything in the tree is found by searching for it based on its key.*/

  __u16 ih_type;		/* direct item, indirect item, sd, dir entry without sd, dir entry with sd, etc */
  __u16 ih_free_space;		/* amount of free space in the last unfm node pointed to by the indirect item */
  __u16 ih_item_len;		/* total size of the item body */
  __u16 ih_item_location;	/* an offset to the item body within the block */
};

#define IH_SIZE (sizeof(struct item_head))


/* there are 4 item types (stat data, direct item, indirect item, dir entry item  */
#define IH_Stat_data 0
#define IH_Indirect 1
#define IH_Direct 2
#define IH_Dir_entry 3

/* dir entry item attributes (can be with stat data, pointing to directory, hidden)) */
#define IH_Dir_entry_with_stat_data 4
#define IH_Dir_entry_points_to_dir 5
/* if the above 2 bits are not set, that entry was created as hard link */
#define IH_Dir_entry_hidden 6

/* these are set only in virtual node */
#define VI_Leftmergeable 7
#define VI_Rightmergeable 8


/* key contains information to define type of corresponding item */
#define mark_key_direct(key) ((key)->group_locale.minor = 1, (key)->k_offset |= 0x80000000)	/* set 31-st bit */
#define mark_key_indirect(key) ((key)->group_locale.minor = 1, (key)->k_offset &= ~0x80000000)	/* clear 31-st bit */
#define mark_key_dir_entry(key) ((key)->k_offset = 0)

#define is_key_of_stat_data(key) ((key)->name_hash == 0 && (key)->k_offset == 0)
#define is_key_of_indirect_item(key) ((key)->k_offset && !((key)->k_offset & 0x80000000))
#define is_key_of_direct_item(key) ((key)->k_offset & 0x80000000)
#define is_key_of_dir_entry(key) ((key)->name_hash != 0 && (key)->k_offset == 0) 


/* only ih_type defines exactly what this item is */
#define set_item_type(ih,type) ((ih)->ih_type = (1 << type))
#define mark_stat_data(ih) set_item_type (ih, IH_Stat_data)
#define mark_direct_item(ih) set_item_type (ih, IH_Direct)
#define mark_indirect_item(ih) set_item_type (ih, IH_Indirect)
#define mark_dir_entry(ih) set_item_type (ih, IH_Dir_entry)

#define mark_entry_with_stat_data(ih) set_bit (IH_Dir_entry_with_stat_data, &((ih)->ih_type))
#define mark_entry_points_to_dir(ih) set_bit (IH_Dir_entry_points_to_dir, &((ih)->ih_type))
#define mark_entry_hidden(ih) set_bit (IH_Dir_entry_hidden, &((ih)->ih_type))


#define is_stat_data(ih) test_bit (IH_Stat_data, &((ih)->ih_type))
#define is_direct_item(ih) test_bit (IH_Direct, &((ih)->ih_type))
#define is_indirect_item(ih) test_bit (IH_Indirect, &((ih)->ih_type))
#define is_dir_entry(ih) test_bit (IH_Dir_entry, &((ih)->ih_type))
#define is_entry_with_stat_data(ih) test_bit (IH_Dir_entry_with_stat_data, &((ih)->ih_type))
#define is_entry_pointing_to_dir(ih) test_bit (IH_Dir_entry_points_to_dir, &((ih)->ih_type))
#define is_entry_hidden(ih) test_bit (IH_Dir_entry_hidden, &((ih)->ih_type))


/* Our function for comparing keys can compare keys of different lengths.  It takes as a parameter
   the length of the keys it is to compare.  These defines are used in determining what is to be
   passed to it as that parameter. */
#define REISERFS_KEY_LEN     4
/* object locale of directory is 1, though */
#define REISERFS_OBJECT_LOCALE_LEN 3
#define OBJECT_LOCALE_SIZE (3 * sizeof (__u32))


/* The result of the key compare */
#define FIRST_GREATER 1
#define SECOND_GREATER -1
#define KEYS_IDENTICAL 0
#define KEY_FOUND 1
#define KEY_NOT_FOUND 0



/* return values for search_by_key and clones */
#define ITEM_FOUND 1
#define ITEM_NOT_FOUND 0


#define ENTRY_FOUND 1
#define ENTRY_NOT_FOUND 0
#define ENTRY_FOUND_HIDDEN 3
#define GOTO_PREVIOUS_ITEM 2



#define DIRECTORY_NOT_FOUND -1
#define BYTE_FOUND 1
#define BYTE_NOT_FOUND 0
#define FILE_NOT_FOUND -1

#define POSITION_FOUND 1
#define POSITION_NOT_FOUND 0
#define POSITION_FOUND_INVISIBLE 3




/* get stat-data */
#define B_I_STAT_DATA(bh, ih) ( (struct stat_data * )((bh)->b_data + (ih)->ih_item_location) )


#define I_K_KEY_IN_ITEM(p_s_ih, p_s_key, n_blocksize) \
    ( ! items_of_different_files(&((p_s_ih)->ih_key), p_s_key) && \
          I_OFF_BYTE_IN_ITEM(p_s_ih, (p_s_key)->k_offset, n_blocksize) )

/* maximal length of item */ 
#define MAX_ITEM_LEN(block_size) (block_size - BLKH_SIZE - IH_SIZE)

/* object identifier for root dir */
#define REISERFS_ROOT_OBJECTID 2
#define REISERFS_ROOT_PARENT_OBJECTID 1
extern struct key root_key;




/* 
 * Picture represents a leaf of the S+tree
 *  ______________________________________________________
 * |      |  Array of     |                   |           |
 * |Block |  Object-Item  |      F r e e      |  Objects- |
 * | head |  Headers      |     S p a c e     |   Items   |
 * |______|_______________|___________________|___________|
 */

struct block_head {                /* Header of a disk block.  More precisely, header of a formatted leaf or
				      internal node, and not the header of an unformatted node. */
  unsigned short blk_level;        /* Level of a block in the tree. */
  unsigned short blk_nr_item;      /* Number of keys/items in a block. */

  unsigned short blk_free_space;   /* Block free space in bytes.  */
  struct key  blk_right_delim_key; /* Right delimiting key for this block (supported for leaf level nodes
				      only) */
};

#define BLKH_SIZE (sizeof(struct block_head))

/*
 * values for blk_type field
 */

#define FREE_LEVEL        0 /* Node of this level is out of the tree. */

#define DISK_LEAF_NODE_LEVEL  1 /* Leaf node level.                       */

/* Given the buffer head of a formatted node, resolve to the block head of that node. */
#define B_BLK_HEAD(p_s_bh)  ((struct block_head *)((p_s_bh)->b_data))
/* Number of items that are in buffer. */
#define B_NR_ITEMS(p_s_bh)	  	( B_BLK_HEAD(p_s_bh)->blk_nr_item )

/* Get right delimiting key. */
#define B_PRIGHT_DELIM_KEY(p_s_bh)	( &(B_BLK_HEAD(p_s_bh)->blk_right_delim_key) )

/* Does the buffer contain a disk leaf. */
#define B_IS_ITEMS_LEVEL(p_s_bh)   	( B_BLK_HEAD(p_s_bh)->blk_level == DISK_LEAF_NODE_LEVEL )

/* Does the buffer contain a disk internal node */
#define B_IS_KEYS_LEVEL(p_s_bh) 	( B_BLK_HEAD(p_s_bh)->blk_level > DISK_LEAF_NODE_LEVEL )



/***************************************************************************/
/*                             STAT DATA                                   */
/***************************************************************************/

/* Common stat data for every object (reiserfs version of UFS disk inode minus the address blocks) */
struct stat_data {
  __u16 sd_mode;	/* file type, permissions */
  __u16 sd_nlink;	/* number of hard links */
  __u16 sd_uid;		/* owner */
  __u16 sd_gid;		/* group */
  __u32 sd_size;	/* file size */
  __u32 sd_atime;	/* time of last access */
  __u32 sd_mtime;	/* time file was last modified  */
  __u32 sd_ctime;	/* time inode (stat data) was last changed (except changes to sd_atime and sd_mtime) */
  union {
    __u32 first_direct_byte;
    __u32 parent_dir_id;
    __u32 rdev;
  } u;
};

#define SD_SIZE (sizeof(struct stat_data))

/* empty directory contains one item and it is directory stat data item */
#define EMPTY_DIR_SIZE  (IH_SIZE + SD_SIZE)


/***************************************************************************/
/*                      DIRECTORY ENTRY STRUCTURE                          */
/***************************************************************************/
/* Entry is stored in the separate item. This items can contains a stat data of regular file or symlink
   Picture represents the structure of directory entry:
   --------------------------------------
   |   stat_data  |  nothing     | name |	- reiserfs_create or reiserfs_symlink or ...
   or
   |   nothing    | object_locale| name |	- reiserfs_link
   or
   |   nothing    | nothing      | name |	- reiserfs_mkdir
   --------------------------------------

   If entry points to directory, nothing but name is stored in the entry
   When stat data is in entry, object_locale is not stored in the
   entry. And if there is no stat data in the entry, key is stored in it.

   we do not store entry "." and ".." like other entries. All needed information about those entries
   can be taken from directory stat data.
*/


#define offset_to_name(ih) ((is_entry_with_stat_data (ih)) ? (SD_SIZE) : \
(is_entry_pointing_to_dir (ih) ? 0 : OBJECT_LOCALE_SIZE))


/* name in entry item */
#define B_I_NAME(bh,ih) (B_I_PITEM (bh, ih) + offset_to_name (ih))

/* length of name in entry */
#define B_I_NAMELEN(bh,ih) ((ih)->ih_item_len - offset_to_name (ih))

/* key of referenced object. Item is assumed to be a directory entry item not pointing to directory */
#define B_I_OBJECT_LOCALE_FROM_ENTRY(bh,ih) \
((is_entry_with_stat_data (ih)) ? &(ih->ih_key) : (struct key *)B_I_PITEM (bh, ih))


/* it is limited by block size only. This makes a possibility of adding two new leaves in one
   balancing which is quite complex. If it was less (half of block) we would do not need to support
   such balancing */
#define REISERFS_MAX_NAME_LEN(block_size) ((block_size) - BLKH_SIZE - IH_SIZE - KEY_SIZE - SD_SIZE)


/* hash value occupies 24 bits starting from 7 up to 30 */
#define GET_HASH_VALUE(offset) ((offset) & 0x7fffff80)
/* generation number occupies 7 bits starting from 0 up to 6 */
#define GET_GENERATION_NUMBER(offset) ((offset) & 0x0000007f)


/*
 * Picture represents an internal node of the reiserfs tree
 *  ______________________________________________________
 * |      |  Array of     |  Array of         |  Free     |
 * |block |    keys       |  pointers         | space     |
 * | head |      N        |      N+1          |           |
 * |______|_______________|___________________|___________|
 */

/***************************************************************************/
/*                      DISK CHILD                                         */
/***************************************************************************/
/* Disk child pointer: The pointer from an internal node of the tree
   to a node that is on disk. */
struct disk_child {
  unsigned long       dc_block_number;              /* Disk child's block number. */
  unsigned short      dc_size;		            /* Disk child's used space.   */
};

#define DC_SIZE (sizeof(struct disk_child))

/* Get disk child by buffer header and position in the tree node. */
#define B_N_CHILD(p_s_bh,n_pos)  ((struct disk_child *)\
((p_s_bh)->b_data+BLKH_SIZE+B_NR_ITEMS(p_s_bh)*KEY_SIZE+DC_SIZE*(n_pos)))

/* Get disk child number by buffer header and position in the tree node. */
#define B_N_CHILD_NUM(p_s_bh,n_pos) (B_N_CHILD(p_s_bh,n_pos)->dc_block_number)

 /* maximal value of field child_size in structure disk_child */ 
 /* child size is the combined size of all items and their headers */
#define MAX_CHILD_SIZE(bh) ((int)( (bh)->b_size - BLKH_SIZE ))

/* amount of used space in buffer (not including block head) */
#define B_CHILD_SIZE(cur) (MAX_CHILD_SIZE(cur)-(BLK_HEAD(cur)->blk_free_space))

/* max and min number of keys in internal node */
#define MAX_NR_KEY(bh) ( (MAX_CHILD_SIZE(bh)-DC_SIZE)/(KEY_SIZE+DC_SIZE) )
#define MIN_NR_KEY(bh)    (MAX_NR_KEY(bh)/2)


/***************************************************************************/
/*                      PATH STRUCTURES AND DEFINES                        */
/***************************************************************************/


/* Search_by_key fills up the path from the root to the leaf as it descends the tree looking for the
   key.  It uses reiserfs_bread to try to find buffers in the cache given their block number.  If it
   does not find them in the cache it reads them from disk.  For each node search_by_key finds using
   reiserfs_bread it then uses bin_search to look through that node.  bin_search will find the
   position of the block_number of the next node if it is looking through an internal node.  If it
   is looking through a leaf node bin_search will find the position of the item which has key either
   equal to given key, or which is the maximal key less than the given key. */

struct  path_element  {
  struct buffer_head *	pe_buffer;    /* Pointer to the buffer at the path in the tree. */
  int         		pe_position;  /* Position in the tree node which is placed in the */
                                      /* buffer above.                                  */
};

#define MAX_HEIGHT 5 /* maximal height of a tree */
#define EXTENDED_MAX_HEIGHT         7 /* Must be equals MAX_HEIGHT + FIRST_PATH_ELEMENT_OFFSET */
#define FIRST_PATH_ELEMENT_OFFSET   2 /* Must be equal to at least 2. */

#define ILLEGAL_PATH_ELEMENT_OFFSET 1 /* Must be equal to FIRST_PATH_ELEMENT_OFFSET - 1 */
#define MAX_FEB_SIZE 6   /* this MUST be MAX_HEIGHT + 1. See about FEB below */



/* We need to keep track of who the ancestors of nodes are.  When we
   perform a search we record which nodes were visited while
   descending the tree looking for the node we searched for. This list
   of nodes is called the path.  This information is used while
   performing balancing.  Note that this path information may become
   invalid, and this means we must check it when using it to see if it
   is still valid. You'll need to read search_by_key and the comments
   in it, especially about decrement_counters_in_path(), to understand
   this structure. */
struct  path {
  struct  path_element  path_elements[EXTENDED_MAX_HEIGHT];	/* Array of the path elements.  */
  int                   path_length;                      	/* Position of last filled path element, not number of filled array elements */
  int			pos_in_item;
};

/* Get path element by path and path position. */
#define PATH_OFFSET_PELEMENT(p_s_path,n_offset)  ((p_s_path)->path_elements +(n_offset))

/* Get buffer header at the path by path and path position. */
#define PATH_OFFSET_PBUFFER(p_s_path,n_offset)   (PATH_OFFSET_PELEMENT(p_s_path,n_offset)->pe_buffer)

/* Get position in the element at the path by path and path position. */
#define PATH_OFFSET_POSITION(p_s_path,n_offset) (PATH_OFFSET_PELEMENT(p_s_path,n_offset)->pe_position)


#define PATH_PLAST_BUFFER(p_s_path) (PATH_OFFSET_PBUFFER((p_s_path), (p_s_path)->path_length))
#define PATH_LAST_POSITION(p_s_path) (PATH_OFFSET_POSITION((p_s_path), (p_s_path)->path_length))


#define PATH_PITEM_HEAD(p_s_path)    B_N_PITEM_HEAD(PATH_PLAST_BUFFER(p_s_path),PATH_LAST_POSITION(p_s_path))

/* in do_balance leaf has h == 0 in contrast with path structure,
   where root has level == 0. That is why we need these defines */
#define PATH_H_PBUFFER(p_s_path, h) PATH_OFFSET_PBUFFER (p_s_path, p_s_path->path_length - (h))	/* tb->S[h] */
#define PATH_H_PPARENT(path, h) PATH_H_PBUFFER (path, (h) + 1)			/* tb->F[h] or tb->S[0]->b_parent */
#define PATH_H_POSITION(path, h) PATH_OFFSET_POSITION (path, path->path_length - (h))	
#define PATH_H_B_ITEM_ORDER(path, h) PATH_H_POSITION(path, h + 1)		/* tb->S[h]->b_item_order */

#define PATH_H_PATH_OFFSET(p_s_path, n_h) ((p_s_path)->path_length - (n_h))


/***************************************************************************/
/*                       MISC                                              */
/***************************************************************************/


/* Size of pointer to the unformatted node. */
#define UNFM_P_SIZE (sizeof(__u32))

/* key of first item of the object */
#define INODE_PKEY(inode) ((struct key *)((inode)->u.reiserfs_i.i_key))

/* directory has fair stat data, that is it is not joined with directory entry. Use those fields to
   store key that is stored in ".." */
#define INODE_PARENT_DIR_ID(inode) ((inode)->u.reiserfs_i.i_parent_dir_id)



#ifdef REISERFS_CHECK
extern int g_kmalloc_count;
#endif




/***************************************************************************/
/*                PRESERVE LIST STUFF                                      */
/***************************************************************************/
/* This flag tracks whether a buffer might contain data that has been shifted to it from another
   node which is on disk and might be obliterated if power fails and this buffer is not written to
   disk before that other node is. This flag is cleared in unlock buffers..  Note that this assumes
   that write_caching is turned off for the disk.  Linux drivers turn write_caching off by default,
   so this should be correct.  I need this flag to ensure that we don't pass a block, from which
   items were shifted, to the scsi controller, free the preserve list, use a freed block, and then
   have the scsi controller reorder the writes so that the supposedly preserved block is overwritten
   before the block containing the shifted items reaches disk, and then risk the system crashing
   before the shifted items reach disk.  See preserve.c for a discussion of the preserve list and
   its role. -Hans */
#define BH_Suspected_Recipient	12

/* 1 if the node that this buffer has assigned to it has not been written to disk since it was
   assigned to the buffer.  If a buffer is BH_Unwritten then there is no need to preserve the
   contents of its block after balancing shifts data from it.  See preserve.c for a discussion of
   the preserve list and its role. */
#define BH_Unwritten	13

/* 1 if the block that this buffer was last read from or written to has been placed on the preserved
   list.  See preserve.c for a discussion of the preserve list and its role. */
#define BH_Preserved	14


/* modes of preserving */
#define PRESERVE_DIRECT_TO_INDIRECT 1
#define PRESERVE_INDIRECT_TO_DIRECT 2
#define PRESERVE_RENAMING 3
#define NOTHING_SPECIAL 4


/* return value for get_space_from_preserve_list */
#define PRESERVE_LIST_WAS_EMPTY 0
#define FEW_BLOCKS_ARE_FREED 1


#define MAX_UL_INT 0xffffffff
#define MAX_INT    0x7ffffff
#define MAX_US_INT 0xffff

#define MAX_KEY_OFFSET		MAX_UL_INT
#define MAX_KEY_UNIQUENESS	MAX_UL_INT
#define MAX_KEY_OBJECTID	MAX_UL_INT

#define MAX_B_NUM  MAX_UL_INT
#define MAX_FC_NUM MAX_US_INT


/* the purpose is to detect overflow of an unsigned short */
#define REISERFS_LINK_MAX (MAX_US_INT - 1000)


/* The following defines are used in reiserfs_insert_item and reiserfs_append_item  */
#define REISERFS_KERNEL_MEM		0	/* reiserfs kernel memory mode	*/
#define REISERFS_USER_MEM		1	/* reiserfs user memory mode		*/


/***************************************************************************/
/*                  FIXATE NODES                                           */
/***************************************************************************/


/* To make any changes in the tree we always first find node, that contains item to be changed/deleted or
   place to insert a new item. We call this node S. To do balancing we need to decide what we will shift to
   left/right neighbor, or to a new node, where new item will be etc. To make this analysis simpler we build
   virtual node. Virtual node is an array of items, that will replace items of node S. (For instance if we are
   going to delete an item, virtual node does not contain it). Virtual node keeps information about item sizes
   and types, mergeability of first and last items, sizes of all entries in directory item. We use this array
   of items when calculating what we can shift to neighbors and how many nodes we have to have if we do not
   any shiftings, if we shift to left/right neighbor or to both. */
struct virtual_item
{
  unsigned short vi_type;		/* item type, mergeability */
  unsigned short vi_item_len;           /* length of item that it will have after balancing */
  
  short vi_entry_count;			/* number of entries in directory item (including the new one if any,
					   or excluding entry if it must be cut) */
  unsigned short * vi_entry_sizes;	/* array of entry lengths for directory item */
};

struct virtual_node
{
  char * vn_free_ptr;		/* this is a pointer to the free space in the buffer */
  unsigned short vn_nr_item;	/* number of items in virtual node */
  short vn_size;        	/* size of node , that node would have if it has unlimited size and no balancing is performed */
  short vn_mode;		/* mode of balancing (paste, insert, delete, cut) */
  short vn_affected_item_num; 
  short vn_pos_in_item;
  struct item_head * vn_ins_ih;	/* item header of inserted item, 0 for other modes */
  struct virtual_item * vn_vi;	/* array of items (including a new one, excluding item to be deleted) */
};


/***************************************************************************/
/*                  TREE BALANCE                                           */
/***************************************************************************/

/* This temporary structure is used in tree balance algorithms, and
   constructed as we go to the extent that its various parts are
   needed.  It contains arrays of nodes that can potentially be
   involved in the balancing of node S, and parameters that define how
   each of the nodes must be balanced.  Note that in these algorithms
   for balancing the worst case is to need to balance the current node
   S and the left and right neighbors and all of their parents plus
   create a new node.  We implement S1 balancing for the leaf nodes
   and S0 balancing for the internal nodes (S1 and S0 are defined in
   our papers.)*/

#define MAX_FREE_BLOCK 7	/* size of the array of buffers to free at end of do_balance */

/*#define MAX_DIRTIABLE 3*/		/* L, S, R */
#define MAX_PRESERVE_NODES 2

/* maximum number of FEB blocknrs on a single level */
#define MAX_AMOUNT_NEEDED 2

/* someday somebody will prefix every field in this struct with tb_ */
struct tree_balance
{
  struct super_block * tb_sb;
  struct path * tb_path;
  struct buffer_head * L[MAX_HEIGHT];        /* array of left neighbors of nodes in the path */
  struct buffer_head * R[MAX_HEIGHT];        /* array of right neighbors of nodes in the path*/
  struct buffer_head * FL[MAX_HEIGHT];       /* array of fathers of the left  neighbors      */
  struct buffer_head * FR[MAX_HEIGHT];       /* array of fathers of the right neighbors      */
  struct buffer_head * CFL[MAX_HEIGHT];      /* array of common parents of center node and its left neighbor  */
  struct buffer_head * CFR[MAX_HEIGHT];      /* array of common parents of center node and its right neighbor */

 /* array of blocknr's that are free and are the nearest to the left node that are usable
    for writing dirty formatted leaves, using the write_next_to algorithm. */
  /*unsigned long free_and_near[MAX_DIRTIABLE];*/

  /* nodes will be used in preserving */
                                /* so we don't just get blocks, we have to get buffers which we likely won't use.  Seems
                                   like it is not optimal. -Hans */
  struct buffer_head * tb_nodes_for_preserving[MAX_PRESERVE_NODES];

  struct buffer_head * FEB[MAX_FEB_SIZE]; /* array of empty buffers. Number of buffers in array equals
					     cur_blknum. */

 short int lnum[MAX_HEIGHT];	/* array of number of items which must be shifted to the left in
				   order to balance the current node; for leaves includes item
				   that will be partially shifted; for internal nodes, it is
				   the number of child pointers rather than items. It includes
				   the new item being created.  For preserve_shifted() purposes
				   the code sometimes subtracts one from this number to get the
				   number of currently existing items being shifted, and even
				   more often for leaves it subtracts one to get the number of
				   wholly shifted items for other purposes. */
 short int rnum[MAX_HEIGHT];	/* substitute right for left in comment above */
 short int lkey[MAX_HEIGHT];               /* array indexed by height h mapping the key delimiting L[h] and
					      S[h] to its item number within the node CFL[h] */
 short int rkey[MAX_HEIGHT];               /* substitute r for l in comment above */
 short int insert_size[MAX_HEIGHT];        /* the number of bytes by we are trying to add or remove from
					      S[h]. A negative value means removing.  */
 short int blknum[MAX_HEIGHT];             /* number of nodes that will replace node S[h] after
					      balancing on the level h of the tree.  If 0 then S is
					      being deleted, if 1 then S is remaining and no new nodes
					      are being created, if 2 or 3 then 1 or 2 new nodes is
					      being created */

/* fields that are used only for balancing leaves of the tree */
 short int cur_blknum;	/* number of empty blocks having been already allocated			*/
 short int s0num;             /* number of items that fall into left most  node when S[0] splits	*/
 short int s1num;             /* number of items that fall into first  new node when S[0] splits	*/
 short int s2num;             /* number of items that fall into second new node when S[0] splits	*/
 short int lbytes;            /* number of bytes which can flow to the left neighbor from the	left	*/
                              /* most liquid item that cannot be shifted from S[0] entirely		*/
                              /* if -1 then nothing will be partially shifted */
 short int rbytes;            /* number of bytes which will flow to the right neighbor from the right	*/
			      /* most liquid item that cannot be shifted from S[0] entirely		*/
                              /* if -1 then nothing will be partially shifted                           */
 short int s1bytes;		/* number of bytes which flow to the first  new node when S[0] splits	*/
            			/* note: if S[0] splits into 3 nodes, then items do not need to be cut	*/
 short int s2bytes;
 struct buffer_head * buf_to_free[MAX_FREE_BLOCK]; /* buffers which are to be freed after do_balance finishes by unfix_nodes */
 char * vn_buf;	/* the buffer header for the virtual node used to calculate node packing */
 int vn_buf_size;
 char preserve_mode;		/* indicates that the deletion that will be done is part of a conversion of
				   a tail to an unformatted node, and the buffer with the item should be
				   preserve_shifted(). see preserve.c */
 } ;


#if 0
				/* when balancing we potentially affect a 3 node wide column of nodes
                                   in the tree (the top of the column may be tapered). C is the nodes
                                   at the center of this column, and L and R are the nodes to the
                                   left and right.  */
  struct seal * L_path_seals[MAX_HEIGHT];
  struct seal * C_path_seals[MAX_HEIGHT];
  struct seal * R_path_seals[MAX_HEIGHT];
  char L_path_lock_types[MAX_HEIGHT];   /* 'r', 'w', or 'n' for read, write, or none */
  char C_path_lock_types[MAX_HEIGHT];
  char R_path_lock_types[MAX_HEIGHT];


  struct seal_list_elem * C_seal[MAX_HEIGHT];        /* array of seals on nodes in the path */
  struct seal_list_elem * L_seal[MAX_HEIGHT];        /* array of seals on left neighbors of nodes in the path */
  struct seal_list_elem * R_seal[MAX_HEIGHT];        /* array of seals on right neighbors of nodes in the path*/
  struct seal_list_elem * FL_seal[MAX_HEIGHT];       /* array of seals on fathers of the left  neighbors      */
  struct seal_list_elem * FR_seal[MAX_HEIGHT];       /* array of seals on fathers of the right neighbors      */
  struct seal_list_elem * CFL_seal[MAX_HEIGHT];      /* array of seals on common parents of center node and its left neighbor  */
  struct seal_list_elem * CFR_seal[MAX_HEIGHT];      /* array of seals on common parents of center node and its right neighbor */
 
  struct char C_desired_lock_type[MAX_HEIGHT]; /* 'r', 'w', or 'n' for read, write, or none */
  struct char L_desired_lock_type[MAX_HEIGHT];        
  struct char R_desired_lock_type[MAX_HEIGHT];        
  struct char FL_desired_lock_type[MAX_HEIGHT];       
  struct char FR_desired_lock_type[MAX_HEIGHT];       
  struct char CFL_desired_lock_type[MAX_HEIGHT];      
  struct char CFR_desired_lock_type[MAX_HEIGHT];      
#endif





/* These are modes of balancing */

/* When inserting an item. */
#define M_INSERT	'i'
/* When inserting into (directories only) or appending onto an already
   existant item. */
#define M_PASTE		'p'
/* When deleting an item. */
#define M_DELETE	'd'
/* When truncating an item or removing an entry from a (directory) item. */
#define M_CUT 		'c'

/* used when balancing on leaf level skipped (in reiserfsck) */
#define M_INTERNAL	'n'

/* When further balancing is not needed, then do_balance does not need
   to be called. */
#define M_SKIP_BALANCING 		's'
#define M_CONVERT	'v'

/* modes of leaf_move_items */
#define LEAF_FROM_S_TO_L 0
#define LEAF_FROM_S_TO_R 1
#define LEAF_FROM_R_TO_L 2
#define LEAF_FROM_L_TO_R 3
#define LEAF_FROM_S_TO_SNEW 4

#define FIRST_TO_LAST 0
#define LAST_TO_FIRST 1

/* used in do_balance for passing parent of node information that has
   been gotten from tb struct */
struct buffer_info {
	struct buffer_head * bi_bh;
	struct buffer_head * bi_parent;
	int bi_position;
};


/* there are 4 types of items: stat data, directory item, indirect, direct.
+-------------------+------------+--------------+------------+
|	            |  k_offset  | k_uniqueness | mergeable? |
+-------------------+------------+--------------+------------+
|     stat data     |	0        |      0       |   no       |
+-------------------+------------+--------------+------------+
| 1st directory item| DOT_OFFSET |DIRENTRY_UNIQUENESS|   no       | 
| non 1st directory | hash value |              |   yes      |
|     item          |            |              |            |
+-------------------+------------+--------------+------------+
| indirect item     | offset + 1 |TYPE_INDIRECT |   if this is not the first indirect item of the object
+-------------------+------------+--------------+------------+
| direct item       | offset + 1 |TYPE_DIRECT   | if not this is not the first direct item of the object
+-------------------+------------+--------------+------------+
*/
#define TYPE_STAT_DATA 0x0
#define TYPE_DIRECT 0xffffffff
#define TYPE_INDIRECT 0xfffffffe
#define TYPE_DIRECTORY_MAX 0xfffffffd


/*
#define KEY_IS_STAT_DATA_KEY(ih) 	( test_bit (IH_Stat_data, &((p_s_key)->k_type)) )
#define KEY_IS_DIRECTORY_KEY(p_s_key)	( test_bit (K_Dir_entry, &((p_s_key)->k_type)) )
#define KEY_IS_DIRECT_KEY(p_s_key) 	( test_bit (K_Direct_item, &((p_s_key)->k_type)) )
#define KEY_IS_INDIRECT_KEY(p_s_key)	( test_bit (K_Indirect_item, &((p_s_key)->k_type)) )
*/

#define I_IS_STAT_DATA_ITEM(ih) (test_bit (IH_Stat_data, &((ih)->ih_type)))
#define I_IS_DIRECTORY_ITEM(ih) (test_bit (IH_Dir_entry,  &((ih)->ih_type)))
#define I_IS_DIRECT_ITEM(ih) 	(test_bit (IH_Direct,  &((ih)->ih_type)))
#define I_IS_INDIRECT_ITEM(ih) 	(test_bit (IH_Indirect,  &((ih)->ih_type)))


#ifdef __KERNEL__


/*
extern inline void copy_short_key (void * to, void * from)
{
  memcpy (to, from, SHORT_KEY_SIZE);
}
*/


#endif	/* __KERNEL__ */



/* number of blocks pointed to by the indirect item */
#define I_UNFM_NUM(p_s_ih)	( (p_s_ih)->ih_item_len / UNFM_P_SIZE )

/* the used space within the unformatted node corresponding to pos within the item pointed to by ih */
#define I_POS_UNFM_SIZE(ih,pos,size) (((pos) == I_UNFM_NUM(ih) - 1 ) ? (size) - (ih)->ih_free_space : (size))

/* number of bytes contained by the direct item or the unformatted nodes the indirect item points to */

#define I_BYTES_NUMBER(ih,size) (( I_IS_INDIRECT_ITEM(ih) ) ?\
				 I_UNFM_NUM(ih)*size - (ih)->ih_free_space :\
				 (( I_IS_DIRECT_ITEM(ih) ) ? (ih)->ih_item_len : 0 /* stat data */)) 

/* check whether byte number 'offset' is in this item */
#define I_OFF_BYTE_IN_ITEM(p_s_ih, n_offset, n_blocksize) \
                  ( (p_s_ih)->ih_key.k_offset <= (n_offset) && \
                    (p_s_ih)->ih_key.k_offset + I_BYTES_NUMBER(p_s_ih,n_blocksize) > (n_offset) )

/* get the item header */ 
#define B_N_PITEM_HEAD(bh,item_num) ( (struct item_head * )((bh)->b_data + BLKH_SIZE) + (item_num) )

/* get key */
#define B_N_PDELIM_KEY(bh,item_num) ( (struct key * )((bh)->b_data + BLKH_SIZE) + (item_num) )

/* get the key */
#define B_N_PKEY(bh,item_num) ( &(B_N_PITEM_HEAD(bh,item_num)->ih_key) )

/* get item body */
#define B_N_PITEM(bh,item_num) ( (bh)->b_data + B_N_PITEM_HEAD((bh),(item_num))->ih_item_location)

/* get the stat data by the buffer header and the item order */
#define B_N_STAT_DATA(bh,nr) \
( (struct stat_data *)((bh)->b_data+B_N_PITEM_HEAD((bh),(nr))->ih_item_location ) )

                 /* following defines use reiserfs buffer header and item header */
 /* get item body */
#define B_I_PITEM(bh,ih) ( (bh)->b_data + (ih)->ih_item_location )


#define MAX_DIRECT_ITEM_LEN(size) ((size) - BLKH_SIZE - 2*IH_SIZE - SD_SIZE - UNFM_P_SIZE)

/* indirect items consist of entries which contain blocknrs, pos
   indicates which entry, and B_I_POS_UNFM_POINTER resolves to the
   blocknr contained by the entry pos points to */
#define B_I_POS_UNFM_POINTER(bh,ih,pos) (*(((unsigned long *)B_I_PITEM(bh,ih)) + (pos)))

/* Reiserfs buffer cache statistics. */
#ifdef REISERFS_CACHE_STAT
 struct reiserfs_cache_stat
	{
  	int nr_reiserfs_ll_r_block; 		/* Number of block reads. */
  	int nr_reiserfs_ll_w_block; 		/* Number of block writes. */
	int nr_reiserfs_schedule; 		/* Number of locked buffers waits. */
	unsigned long nr_reiserfs_bread;	/* Number of calls to reiserfs_bread function */
	unsigned long nr_returns; /* Number of breads of buffers that were hoped to contain a key but did not after bread completed
				     (usually due to object shifting while bread was executing.)
				     In the code this manifests as the number
				     of times that the repeat variable is nonzero in search_by_key.*/
	unsigned long nr_fixed;		/* number of calls of fix_nodes function */
	unsigned long nr_failed;	/* number of calls of fix_nodes in which schedule occurred while the function worked */
	unsigned long nr_find1;		/* How many times we access a child buffer using its direct pointer from an internal node.*/
	unsigned long nr_find2;	        /* Number of times there is neither a direct pointer to
					   nor any entry in the child list pointing to the buffer. */
	unsigned long nr_find3;	        /* When parent is locked (meaning that there are no direct pointers)
					   or parent is leaf and buffer to be found is an unformatted node. */
	}  cache_stat;
#endif


/***************************************************************************/
/*                    FUNCTION DECLARATIONS                                */
/***************************************************************************/

/* buffer.c */
inline int test_and_wait_on_buffer (struct buffer_head * p_s_bh);
struct buffer_head * reiserfs_getblk (kdev_t n_dev, int n_block, int n_size, int * p_n_repeat);
void reiserfs_rehash_buffer (struct buffer_head * bh, unsigned long block);
void reiserfs_show_buffers (kdev_t dev);
void mark_suspected_recipients_dirty (kdev_t dev);
void invalidate_reiserfs_buffers (kdev_t dev);

/* buffer2.c */
struct buffer_head * reiserfs_bread (kdev_t n_dev, int n_block, int n_size, int * p_n_repeat);
struct buffer_head * reiserfs_get_hash_table (kdev_t  n_dev, int n_block, int n_size, int * p_n_repeat);
int reiserfs_sync_file (struct file * p_s_filp, struct dentry * p_s_dentry);



/* objectid.c */
unsigned long reiserfs_get_unused_objectid (struct super_block * s);
void reiserfs_release_objectid (unsigned long objectid_to_release, struct super_block * s);


/* stree.c */
__u32 offset (struct key * key);
int B_IS_IN_TREE(struct buffer_head *);
void copy_key (void * to, void * from);
extern inline int comp_keys (struct key * k1, struct key * k2);
void copy_object_locale (struct key * to, struct key * from);
void copy_item_head(void * p_v_to, void * p_v_from);
extern inline int  comp_short_keys (struct key * k1, struct key * k2);
int items_of_different_files (struct key * key1, struct key * key2);
int keys_of_different_item_types (struct key * key1, struct key * key2);
int comp_items (struct item_head  * p_s_ih, struct path * p_s_path);
struct key * get_rkey (struct path * p_s_chk_path, struct super_block  * p_s_sb);
inline int bin_search (struct key * p_s_key, void * p_v_base, int p_n_num, int p_n_width, int * p_n_pos);
int search_by_key (struct super_block *, struct key *, struct path *, int * , int, int);
int search_for_position_by_key (struct super_block * p_s_sb, struct key * p_s_key, struct path * p_s_search_path, int * p_n_pos_in_item, int * p_n_repeat);
extern inline void decrement_bcount (struct buffer_head * p_s_bh);
void decrement_counters_in_path (struct path * p_s_search_path);
void init_path (struct path * path);
void pathrelse (struct path * p_s_search_path);
#define OK 0
int reiserfs_insert_item (struct super_block * sb, struct path * path, struct item_head * ih, const char * body, 
			  int zeros_num, int preserve_mode);
int reiserfs_paste_into_item (struct super_block * sb, struct path * path, int * pos_in_item, struct key * key,	
			      const char * body, int paste_size, int memmode, int zeros_number);	/* returns -1 if failed and paste_size otherwise */
int reiserfs_cut_from_item (struct inode * inode, struct super_block * sb, struct path * path, int * pos_in_item, struct key * key,
			    unsigned long new_file_size, int preserve_mode);
int reiserfs_delete_item (struct super_block  * p_s_sb, struct path * p_s_path, int * p_n_pos_in_item, struct key * p_s_item_key,
			  struct buffer_head  * p_s_un_bh, int preserve_mode);

void reiserfs_delete_object (struct inode * p_s_inode);
void reiserfs_truncate_file (struct  inode * p_s_inode);
int lock_inode_to_convert (struct inode * p_s_inode);
void unlock_inode_after_convert (struct inode * p_s_inode);
int increment_i_read_sync_counter (struct inode * p_s_inode);
#define lock_conversion(inode) increment_i_read_sync_counter(inode)
void decrement_i_read_sync_counter (struct inode * p_s_inode);
#define unlock_conversion(inode) decrement_i_read_sync_counter(inode)
int get_buffer_by_range (struct super_block * p_s_sb, struct key * p_s_range_begin, struct key * p_s_range_end, 
			 struct buffer_head ** pp_s_buf, unsigned long * p_n_objectid);
int get_buffers_from_range (struct super_block * p_s_sb, struct key * p_s_range_start, struct key * p_s_range_end,
			    struct buffer_head ** p_s_range_buffers, int n_max_nr_buffers_to_return);

#ifndef REISERFS_FSCK

inline int is_left_mergeable (struct key * key, unsigned long bsize);

#else

int is_left_mergeable (struct super_block * s, struct path * path);
int is_right_mergeable (struct super_block * s, struct path * path);
int are_items_mergeable (struct item_head * left, struct item_head * right, int bsize);

#endif



/* inode.c */
struct inode * reiserfs_iget (struct super_block * s, struct path * path, struct key * key);
void init_item_head (struct item_head * ih, struct inode * inode, __u32 offset, int item_len, int type_bit);
void inode2sd (struct stat_data * sd, struct inode * inode);
void reiserfs_delete_inode (struct inode * inode);
void reiserfs_read_inode (struct inode * inode);
void reiserfs_put_inode (struct inode * inode);
int reiserfs_notify_change(struct dentry * dentry, struct iattr * attr);
int reiserfs_bmap (struct inode * inode, int block);
int reiserfs_readpage (struct file * file, struct page * page);
void reiserfs_write_inode (struct inode * inode);

struct inode * reiserfs_new_inode (const struct inode * dir);
struct inode * reiserfs_new_directory_inode (const struct inode * dir, int mode);
struct inode * reiserfs_new_symlink_inode (const struct inode * dir, const char * symname);

int reiserfs_sync_inode (struct inode * inode);
void if_in_ram_update_sd (struct inode * inode);
void reiserfs_inode_setattr(struct inode * inode,  struct iattr * attr);
void __wait_on_inode(struct inode * inode);

/* namei.c */
int search_by_entry_key (struct super_block * sb, struct key * key, struct path * path, int * pos_in_item, int * repeat);
int reiserfs_lookup (struct inode * dir, struct dentry *dentry);
int reiserfs_create (struct inode * dir, struct dentry *dentry,	int mode);
int reiserfs_mknod (struct inode * dir_inode, struct dentry *dentry, int mode, int rdev);
int reiserfs_mkdir (struct inode * dir, struct dentry *dentry, int mode);
int reiserfs_rmdir (struct inode * dir,	struct dentry *dentry);
int reiserfs_unlink (struct inode * dir, struct dentry *dentry);
int reiserfs_symlink (struct inode * dir, struct dentry *dentry, const char * symname);
int reiserfs_link (struct dentry * old_dentry, struct inode * dir, struct dentry *dentry);
int reiserfs_rename (struct inode * old_dir, struct dentry *old_dentry, struct inode * new_dir, struct dentry *new_dentry);

/* super.c */
void reiserfs_write_super (struct super_block * s);
void reiserfs_put_super (struct super_block * s);
int reiserfs_remount (struct super_block * s, int * flags, char * data);
struct super_block * reiserfs_read_super (struct super_block * s, void * data, int silent);
int reiserfs_statfs (struct super_block * s, struct statfs * buf, int bufsiz);
int init_reiserfs_fs (void);

/* dir.c */
extern struct inode_operations reiserfs_dir_inode_operations;

/* file.c */
extern struct inode_operations reiserfs_file_inode_operations;
int do_direct_to_indirect (struct inode * inode);

/* symlink.c */
extern struct inode_operations reiserfs_symlink_inode_operations;


/* fix_nodes.c */
int fix_nodes (int n_op_mode, struct tree_balance * p_s_tb, int n_pos_in_item, struct item_head * p_s_ins_ih);
void unfix_nodes (struct tree_balance *);
void free_buffers_in_tb (struct tree_balance * p_s_tb);
void decrement_key (struct key *);

/* prints.c */
void reiserfs_panic (struct super_block * s, const char * fmt, ...);
void reiserfs_warning (const char * fmt, ...);
void print_virtual_node (struct virtual_node * vn);
void print_buffer_head (struct buffer_head * bh, char * mes);
void print_indirect_item (struct buffer_head * bh, int item_num);
void print_tb (int mode, int item_pos, int pos_in_item, struct tree_balance * tb, char * mes);
void print_dc (struct disk_child * dc);
void print_bi (struct buffer_info * bi, char * mes);
#define PRINT_LEAF_ITEMS 1   /* print all items */
#define PRINT_DIRECTORY_ITEMS 2 /* print directory items */
#define PRINT_DIRECT_ITEMS 4 /* print contents of direct items */
void print_block (struct buffer_head * bh, int print_mode, int first, int last);
void print_path (struct tree_balance * tb, struct path * path);
void print_bmap (int bmapnr, char ** bmap, int blocksize, char * mes);
void print_super_block (struct super_block * s, char * mes);
void print_objectid_map (struct super_block * s);
void print_block_head (struct buffer_head * bh, char * mes);
/* lbalance.c */
int leaf_move_items (int shift_mode, struct tree_balance * tb, int mov_num, int mov_bytes, struct buffer_head * Snew);
int leaf_shift_left (struct tree_balance * tb, int shift_num, int shift_bytes);
int leaf_shift_right (struct tree_balance * tb, int shift_num, int shift_bytes);
void leaf_delete_items (struct buffer_info * cur_bi, int last_first, int first, int del_num, int del_bytes);
void leaf_insert_into_buf (struct buffer_info * bi, int before, struct item_head * inserted_item_ih, const char * inserted_item_body, 
			   int mem_mode, int zeros_number);
void leaf_paste_in_buffer (struct buffer_info * bi, int pasted_item_num, int pos_in_item, int paste_size, const char * body, int mem_mode, 
			   int zeros_number);
void leaf_cut_from_buffer (struct buffer_info * bi, int cut_item_num, int pos_in_item, int cut_size);
/*void leaf_paste_entries (struct buffer_head * bh, int item_num, int before, int new_entry_count, struct reiserfs_de_head * new_dehs, const char * records,
			 int paste_size);*/

#ifdef __KERNEL__
/* preserve.c */
/*int ready_preserve_list (struct super_block * sb, struct buffer_head * bh, unsigned long * free_and_near, int count);*/
int ready_preserve_list (struct tree_balance *, struct buffer_head * bh);
/*void preserve_shifted (struct super_block * sb, struct buffer_info *, unsigned long free_and_near, struct buffer_head *);*/
void preserve_shifted (struct tree_balance *, struct buffer_head **, struct buffer_head *, int, struct buffer_head *);
void add_to_preserve (unsigned long blocknr, struct super_block * sb);
int maybe_free_preserve_list (struct super_block * sb);
int get_space_from_preserve_list (struct super_block * s);
inline void unpreserve (struct super_block * s, struct buffer_head * bh);
void preserve_invalidate (struct tree_balance * tb, struct buffer_head * bh, struct buffer_head *);
unsigned long free_and_near_block (struct tree_balance * tb);
inline void mark_suspected_recipient (struct super_block * sb, struct buffer_head * bh);
inline void unmark_suspected_recipient (struct super_block * sb, struct buffer_head * bh);
int is_buffer_unwritten (struct buffer_head * bh);
int is_buffer_preserved (struct buffer_head * bh);
int is_buffer_suspected_recipient (struct buffer_head * bh);
inline void mark_buffer_unwritten (struct buffer_head * bh);
void unpreserve (struct super_block * s, struct buffer_head * bh);
#ifdef REISERFS_CHECK
void preserve_trace_init_bitmap (struct super_block * s);
void preserve_trace_release_bitmap (struct super_block * s);
#endif
#endif /* __KERNEL__ */


/* ibalance.c */
int balance_internal (struct tree_balance * , int, int, struct item_head * , struct buffer_head **);

/* do_balance.c */
void do_balance (struct tree_balance * tb, int pos_in_item, struct item_head * ih, const char * body, int flag, int mem_mode, int zeros_num);
void reiserfs_invalidate_buffer (struct tree_balance * tb, struct buffer_head * bh, int);
int get_left_neighbor_position (struct tree_balance * tb, int h);
int get_right_neighbor_position (struct tree_balance * tb, int h);
void replace_key (struct buffer_head *, int, struct buffer_head *, int);
void replace_lkey (struct tree_balance *, int, struct item_head *);
void replace_rkey (struct tree_balance *, int, struct item_head *);
void make_empty_node (struct buffer_info *);
struct buffer_head * get_FEB (struct tree_balance *);

/* bitmap.c */
int is_reusable (struct super_block * s, unsigned long block, int bit_value);
void reiserfs_free_block (struct super_block *, unsigned long);
int reiserfs_new_blocknrs (struct super_block *, unsigned long *, unsigned long, int);

/* teahash3.c */
unsigned long keyed_hash (const char *msg, int len);

/* lock.c */
#ifdef __KERNEL__
inline int try_ulong_lock(unsigned long * lock, unsigned short bitnumber);
void unlock_ulong_lock(unsigned long * lock, unsigned short bitnumber, struct  wait_queue ** l_wait);
#endif

/* log.c */
struct log_block_head {
  __u32 timestamp;
  char sign [16];
};


struct update_inode_record {
  struct key inode_sd_key;	/* reiserfs_update_inode looks for this key */
  struct key sd_item_key;	/* this is found item */

  /* stat data's fields */
  __u16 sd_sd_mode;
  __u32 sd_sd_size;		/* raw object size */

  /* inode's fields */
  __u16 inode_i_mode;
  __u32 inode_i_size;		/* real object size */
  __u32 inode_i_ino;
};


struct create_record {
  __u32 dir_i_ino;	/* new file has been created in this directory */
};


struct new_dir_inode_record {	/* new directory stat data has been successfully inserted */
  __u32 i_ino;	
  struct key sd_key;
  __u32 parent_dir_dirid;
  __u32 parent_dir_objid;
};


struct reiserfs_log_record {
  __u32 log_type;
  union {
    struct update_inode_record update_inode;
    struct new_dir_inode_record new_dir_inode;
    struct create_record create;
  } u;
};

#define REISERFS_LOG_SIZE 1000
#define REISERFS_UPDATE_INODE 0x12345678
#define LOG_SIGN "ReIsErFsLoG,v1"


#ifdef __KERNEL__
int init_log_buffer (struct super_block * s);
void release_log_buffer (struct super_block * s);
void record_update_inode_to_intermediate_buffer (struct inode * inode, struct item_head * ih, struct stat_data * sd);
void flush_log_records (struct super_block * s);
#endif





#ifdef __i386__

/* Thanks to Alexej I. Adamovich for this */
extern __inline__ int 
find_first_nonzero_bit(void * addr, unsigned size) {
  int res;


  if (!size) {
    return (0);
  }

  __asm__("cld\n\t"
	  "xorl %%eax,%%eax\n\t"
	  "repe; scasl\n\t"
	  "je 1f\n\t"
	  "movl -4(%%edi),%%eax\n\t"
	  "subl $4, %%edi\n\t"
	  "bsfl %%eax,%%eax\n\t"
	  "1:\tsubl %%edx,%%edi\n\t"
	  "shll $3,%%edi\n\t"
	  "addl %%edi,%%eax"
	  :"=a" (res)
	  :"c" ((size + 31) >> 5), "D" (addr), "d" (addr)
	  :"cx", "di");

  return (res);
}

#else /* __i386__ */

extern __inline__ int find_next_nonzero_bit(void * addr, unsigned size, unsigned offset)
{
	unsigned int * p = ((unsigned int *) addr) + (offset >> 5);
	unsigned int result = offset & ~31UL;
	unsigned int tmp;

	if (offset >= size)
		return size;
	size -= result;
	offset &= 31UL;
	if (offset) {
		tmp = *p++;
		/* set to zero first offset bits */
		tmp &= ~(~0UL >> (32-offset));
		if (size < 32)
			goto found_first;
		if (tmp != 0U)
			goto found_middle;
		size -= 32;
		result += 32;
	}
	while (size >= 32) {
		if ((tmp = *p++) != 0U)
			goto found_middle;
		result += 32;
		size -= 32;
	}
	if (!size)
		return result;
	tmp = *p;
found_first:
found_middle:
	return result + ffs(tmp);
}

#define find_first_nonzero_bit(addr,size) find_next_nonzero_bit((addr), (size), 0)

#endif /* 0 */


#endif /* _LINUX_REISER_FS_H */


